/*
program:   orx_util.rex
type:      Object REXX, REXXSAA 6.0
purpose:   collection of useful routines for Object REXX programs
version:   0.1.2
date:      1996-
changed:   2005-05-29, renamed "cmd" to "rex"

author:    Rony G. Flatscher
           Rony.Flatscher@wu-wien.ac.at
           (Wirtschaftsuniversitaet Wien, University of Economics and Business
           Administration, Vienna/Austria/Europe)
needs:     ---

usage:     insert a line "::REQUIRES orx_util" or "CALL orx_util"

comments:  - for performance considerations USE ARG is used
             - object itself is accessible, no need to have ORX produce a
               string rendering of the object
           - not finished yet

todo:      handle line comments (--), alternative continuation character (-) in addition to (,)

All rights reserved, copyrighted 1995, 1996 by the author.


*/



/******************************************************************************/
/*                                                                            */
/* name:    make_token_list( line, token_position, bWorkOnFile )              */
/*                                                                            */
/* purpose: extracts Rexx-tokens, i.e. string, constant symbol, delimiters,   */
/*          *and* comments                                                    */
/*                                                                            */
/* returns: array of array of ( token, tokentype, next token_position,        */
/*                              char immediately following, line# )           */
/*                                                                            */
/*          token_type:                                                       */
/*                                                                            */
/*          C ... Comment                                                     */
/*          S ... Symbol                                                      */
/*          Q ... Quoted String                                               */
/*          O ... Operator                                                    */
/*          M ... Miscellaneous, like ();.:[] etc.                            */
/*          '' ... nothing left (whitespace only)                             */
/*                                                                            */
/* created: rgf, 96-02                                                        */


.local ~ orx.WhiteSpace = " " || D2C( 13 ) /* CR */ || D2C( 10 ) /* LF */ || ,
                                 D2C(  9 ) /* HT */ || D2C( 11 ) /* VT */ || ,
                                 D2C( 12 ) /* FF */

.local ~ orx.CRLF       = D2C( 10 ) /* LF */

/*
/* bug with ORX in that it adds a CR *automatically* if a LF is encountered     */
.local ~ orx.CRLF       = D2C( 13 ) /* CR */ || D2C( 10 ) /* LF */
*/

.local ~ orx.BreakLinesAt = 120 /* maximum characters per line          */
.local ~ orx.WidthForXRefs = 90 /* maximum characters per line for generating X-references      */
.local ~ orx.bIndent       = .true      /* pretty print: do indentations ?              */
.local ~ orx.bInsertLines  = .true      /* pretty print: insert empty lines ?           */
.local ~ orx.sIndent       = "   "      /* indentation string                           */
.local ~ orx.sIndentComment= "   "      /* indentation for comments                     */


.local ~ orx.OperatorChars = "~+-\/%*|&="/* not operator: */"><"
.local ~ orx.SymbolChars   = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.?!_"
.local ~ orx.keywords      = .directory ~ new
.local ~ orx.directives    = .directory ~ new
.local ~ orx.functions     = .directory ~ new


/* determine case of keywords */
tmpArray = .array ~ of(  "ADDRESS",,
                         "ADDRESS VALUE",,
                         "ARG",,
                         "CALL",,
                         "CALL ANY",,
                         "CALL ERROR",,
                         "CALL FAILURE",,
                         "CALL HALT",,
                         "CALL NAME",,
                         "CALL NOTREADY",,
                         "CALL OFF",,
                         "CALL ON",,
                         "CALL USER",,
                         "DO",,
                         "DO BY",,
                         "DO FOR",,
                         "DO FOREVER",,
                         "DO OVER",,
                         "DO TO",,
                         "DO UNTIL",,
                         "DO WHILE",,
                         "DROP",,
                         "ELSE",,
                         "END",,
                         "EXIT",,
                         "EXPOSE",,
                         "FORWARD",,
                         "FORWARD ARGUMENTS",,
                         "FORWARD ARRAY",,
                         "FORWARD CLASS",,
                         "FORWARD CONTINUE",,
                         "FORWARD MESSAGE",,
                         "FORWARD TO",,
                         "GUARD",,
                         "GUARD OFF",,
                         "GUARD ON",,
                         "GUARD WHEN",,
                         "IF",,
                         "IF THEN",,
                         "INTERPRET",,
                         "ITERATE",,
                         "LEAVE",,
                         "NOP",,
                         "NUMERIC",,
                         "NUMERIC DIGITS",,
                         "NUMERIC ENGINEERING",,
                         "NUMERIC FORM",,
                         "NUMERIC FUZZ",,
                         "NUMERIC SCIENTIFIC",,
                         "NUMERIC VALUE",,
                         "OPTIONS",,
                         "OTHERWISE",,
                         "PARSE",,
                         "PARSE ARG",,
                         "PARSE CASELESS",,
                         "PARSE LINEIN",,
                         "PARSE LOWER",,
                         "PARSE PULL",,
                         "PARSE SOURCE",,
                         "PARSE UPPER",,
                         "PARSE VALUE",,
                         "PARSE VAR",,
                         "PARSE VERSION",,
                         "PARSE WITH",,
                         "PROCEDURE",,
                         "PROCEDURE EXPOSE",,
                         "PULL",,
                         "PUSH",,
                         "QUEUE",,
                         "RAISE",,
                         "RAISE ADDITIONAL",,
                         "RAISE ARRAY",,
                         "RAISE DESCRIPTION",,
                         "RAISE ERROR",,
                         "RAISE EXIT",,
                         "RAISE FAILURE",,
                         "RAISE HALT",,
                         "RAISE LOSTDIGITS",,
                         "RAISE NOMETHOD",,
                         "RAISE NOSTRING",,
                         "RAISE NOTREADY",,
                         "RAISE NOVALUE",,
                         "RAISE PROPAGATE",,
                         "RAISE RETURN",,
                         "RAISE SYNTAX",,
                         "RAISE USER",,
                         "REPLY",,
                         "RETURN",,
                         "SAY",,
                         "SELECT",,
                         "SELECT OTHERWISE",,
                         "SELECT THEN",,
                         "SELECT WHEN",,
                         "SIGNAL",,
                         "SIGNAL ANY",,
                         "SIGNAL ERROR",,
                         "SIGNAL FAILURE",,
                         "SIGNAL HALT",,
                         "SIGNAL LOSTDIGITS",,
                         "SIGNAL NAME",,
                         "SIGNAL NOMETHOD",,
                         "SIGNAL NOSTRING",,
                         "SIGNAL NOTREADY",,
                         "SIGNAL NOVALUE",,
                         "SIGNAL OFF",,
                         "SIGNAL ON",,
                         "SIGNAL SYNTAX",,
                         "SIGNAL USER",,
                         "SIGNAL VALUE",,
                         "THEN",,
                         "TRACE",,
                         "TRACE ALL",,
                         "TRACE COMMANDS",,
                         "TRACE ERROR",,
                         "TRACE FAILURE",,
                         "TRACE INTERMEDIATES",,
                         "TRACE LABELS",,
                         "TRACE NORMAL",,
                         "TRACE OFF",,
                         "TRACE RESULTS",,
                         "TRACE VALUE",,
                         "USE",,
                         "USE ARG",,
                         "WHEN",,
                         "WHEN THEN",
                       )

DO keyword OVER tmpArray
   .orx.keywords ~ setentry( keyword, WORD( keyword, WORDS( keyword ) ) )
END


/* determine case of directives */
tmpArray = .array ~ of( "CLASS",,
                        "CLASS EXTERNAL",,
                        "CLASS INHERIT",,
                        "CLASS METACLASS",,
                        "CLASS MIXINCLASS",,
                        "CLASS PUBLIC",,
                        "CLASS SUBCLASS",,
                        "METHOD",,
                        "METHOD ATTRIBUTE",,
                        "METHOD CLASS",,
                        "METHOD GUARDED",,
                        "METHOD PRIVATE",,
                        "METHOD UNGUARDED",,
                        "REQUIRES",,
                        "ROUTINE",,
                        "ROUTINE PUBLIC",
                       )

DO keyword OVER tmpArray
   .orx.directives ~ setentry( keyword, WORD( keyword, WORDS( keyword ) ) )
END

/* determine case of "routines" (i.e. procedures, functions) */
tmpArray = .array ~ of( "ABBREV",,
                        "ABS",,
                        "ADDRESS",,
                        "ARG",,
                        "BEEP",,
                        "BITAND",,
                        "BITOR",,
                        "BITXOR",,
                        "B2X",,
                        "CENTER",,
                        "CENTRE",,
                        "CHANGESTR",,
                        "CHARIN",,
                        "CHAROUT",,
                        "CHARS",,
                        "COMPARE",,
                        "CONDITION",,
                        "COPIES",,
                        "COUNTSTR",,
                        "C2D",,
                        "C2X",,
                        "DATATYPE",,
                        "DATE",,
                        "DELSTR",,
                        "DELWORD",,
                        "DIGITS",,
                        "DIRECTORY",,
                        "D2C",,
                        "D2X",,
                        "ERRORTEXT",,
                        "FILESPEC",,
                        "FORM",,
                        "FORMAT",,
                        "FUZZ",,
                        "INSERT",,
                        "LASTPOS",,
                        "LEFT",,
                        "LENGTH",,
                        "LINEIN",,
                        "LINEOUT",,
                        "LINES",,
                        "MAX",,
                        "MIN",,
                        "OVERLAY",,
                        "POS",,
                        "QUEUED",,
                        "RANDOM",,
                        "REVERSE",,
                        "RIGHT",,
                        "SETLOCAL",,
                        "SIGN",,
                        "SOURCELINE",,
                        "SPACE",,
                        "STREAM",,
                        "STRIP",,
                        "SUBSTR",,
                        "SUBWORD",,
                        "SYMBOL",,
                        "TIME",,
                        "TRACE",,
                        "TRANSLATE",,
                        "TRUNC",,
                        "VALUE",,
                        "VAR",,
                        "VERIFY",,
                        "WORD",,
                        "WORDINDEX",,
                        "WORDLENGTH",,
                        "WORDPOS",,
                        "WORDS",,
                        "XRANGE",,
                        "X2B",,
                        "X2C",,
                        "X2D",,
                        "RXFUNCADD",,
                        "RXFUNCDROP",,
                        "RXFUNCQUERY",,
                        "RXQUEUE",,
                        "RxMessageBox",,
                        "SysAddFileHandle",,
                        "SysAddRexxMacro",,
                        "SysBootDrive",,
                        "SysClearRexxMacroSpace",,
                        "SysCloseEventSem",,
                        "SysCloseMutexSem",,
                        "SysCls",,
                        "SysCopyObject",,
                        "SysCreateEventSem",,
                        "SysCreateMutexSem",,
                        "SysCreateObject",,
                        "SysCreateShadow",,
                        "SysCurPos",,
                        "SysCurState",,
                        "SysDeregisterObjectClass",,
                        "SysDestroyObject",,
                        "SysDriveInfo",,
                        "SysDriveMap",,
                        "SysDropFuncs",,
                        "SysDropRexxMacro",,
                        "SysElapsedTime",,
                        "SysFileDelete",,
                        "SysFileSearch",,
                        "SysFileSystemType",,
                        "SysFileTree",,
                        "SysGetCollate",,
                        "SysGetEA",,
                        "SysGetKey",,
                        "SysGetMessage",,
                        "SysIni",,
                        "SysLoadFuncs",,
                        "SysLoadRexxMacroSpace",,
                        "SysMapCase",,
                        "SysMkDir",,
                        "SysMoveObject",,
                        "SysNationalLanguageCompare",,
                        "SysOpenEventSem",,
                        "SysOpenMutexSem",,
                        "SysOpenObject",,
                        "SysOS2Ver",,
                        "SysPostEventSem",,
                        "SysProcessType",,
                        "SysPutEA",,
                        "SysQueryClassList",,
                        "SysQueryEAList",,
                        "SysQueryProcessCodePage",,
                        "SysQueryRexxMacro",,
                        "SysRegisterObjectClass",,
                        "SysReleaseMutexSem",,
                        "SysReorderRexxMacro",,
                        "SysRequestMutexSem",,
                        "SysResetEventSem",,
                        "SysRmDir",,
                        "SysSaveObject",,
                        "SysSaveRexxMacroSpace",,
                        "SysSearchPath",,
                        "SysSetFileHandle",,
                        "SysSetIcon",,
                        "SysSetObjectData",,
                        "SysSetPriority",,
                        "SysSetProcessCodePage",,
                        "SysShutDownSystem",,
                        "SysSleep",,
                        "SysTempFileName",,
                        "SysTextScreenRead",,
                        "SysTextScreenSize",,
                        "SysWaitEventSem",,
                        "SysWaitNamedPipe",,
                        "SysWildCard",
                       )

DO keyword OVER tmpArray
   .orx.functions ~ setentry( keyword, WORD( keyword, WORDS( keyword ) ) )
END

DROP tmpArray           /* indicate variable not needed anymore                 */

:: requires RGF_UTIL.REX /* sorting, remove_quotes */

:: routine pp ; return "[" || arg( 1 ) || "]"


/* writes a pretty print version of the given file to .stdout                   */
:: ROUTINE PrettyPrint PUBLIC
   USE ARG someFile, flag

   IF \VAR( "flag" ) THEN flag = .true          /* default to filename          */
   anArray = make_token_list( someFile, .true ) /* indicate a filename          */
   .stdout ~ arrayout( reassemble( anArray, flag))      /* strip comments ?     */


/* writes a pretty print version of the given file to .stdout                   */
:: ROUTINE CrossReferenceExtended PUBLIC
   USE ARG someFile
   anArray = make_token_list( someFile, .true ) /* indicate a filename          */
   call xref_diff anArray

/* writes a pretty print version of the given file to .stdout                   */
:: ROUTINE CrossReference PUBLIC
   USE ARG someFile
   anArray = make_token_list( someFile, .true ) /* indicate a filename          */
   call xref anArray


/* reassemble the program from the token list, returns pretty printed program in array form     */
:: ROUTINE reassemble           PUBLIC
   USE ARG anArray, bShowComments

   IF \ VAR( "bShowComments" ) THEN bShowComments = .true /* default to show comments                     */
   OutIndex = 1
   OutArray = .array ~ new

   tmpLine = ""
   oldLine = anArray[ 1 ][ 4 ]          /* get very first line number           */
   oldType = ""                         /* old token type                       */
   commNChar = ""                       /* char following comment               */
   arrEntries = anArray ~ items         /* nr. of entries in array              */
   bPrintLine = .true                   /* ready to print line ?                */

   bPrintEmptyLines = .orx.bInsertLines /* print empty lines resp. insert them? */
   bIndent      = .orx.bIndent          /* indent ?                             */
   sIndent      = .orx.sIndent          /* string to use as indentation         */
   sCommIndent  = .orx.sIndentComment   /* string to use for comment indentation*/

   indent       = 0                     /* nr. of indentations at present       */
   nextIndent   = 0                     /* nr. of indentations after present line was printed   */
   PendingDo    = 0                     /* nr of pending do's                   */

   bOneTimeIndention = .false           /* a one time indentation pending ?     */
   bInitialCommentPrinted = .false
   oldLNr       = 0
   bCommPrinted = .false                /* comment printed already ?            */
   bEmptyLinePrinted = .false           /* an empty line was printed previously */
   oldFirstWord = ""
   nextLineBreak = .orx.BreakLinesAt

   DO itemNr = 1 TO arrEntries
      tmpArr = anArray[ itemNr ]        /* get token-array      */
      token = tmpArr[ 1 ]
      tType = tmpArr[ 2 ]
      nChar = tmpArr[ 3 ]
      lNr   = tmpArr[ 4 ]

      IF tType = "N" then               /* a new line arrived                   */
      DO
         bCommPrinted = .false
         IF tmpLine = "" THEN ITERATE

         IF bIndent THEN
         DO
            IF indent > 0 THEN
               tmpLine = COPIES( sIndent, indent ) || tmpLine
         END

         indent  = nextIndent

         IF bOneTimeIndention THEN   /* if an IF or ELSE             */
         DO
            firstWord = TRANSLATE( WORD( tmpLine, 1 ) )
            lastWord = TRANSLATE( WORD( tmpLine, WORDS( tmpLine ) ) )
            oldFirstWord = firstWord            /* remember IF/ELSE, to check with DO/SELECT later on   */

/* ATTENTION: case of "else | then a = do" is ignored in this case ! */
            IF lastWord = "DO" THEN
            DO
               PendingDo  = PendingDo + 1
            END
            ELSE IF ( firstWord = "IF" | firstWord = "WHEN" ) & lastWord <> "THEN" THEN     /* IF ... THEN a = b */
            DO
               nextIndent = nextIndent - 1
               indent = nextIndent
               oldFirstWord = ""
            END
            ELSE IF ( firstWord = "ELSE" | firstWord = "THEN" ) & WORDS( tmpLine ) > 1 & lastWord <> "THEN" THEN /* "ELSE a = b" */
            DO
               nextIndent = nextIndent - 1
               indent = nextIndent
               oldFirstWord = ""
            END
            ELSE IF firstWord = "OTHERWISE" & WORDS( tmpLine ) > 1 THEN
            DO
               nextIndent = nextIndent - 1
               indent = nextIndent
            END
            ELSE        /* IF ... THEN; and ELSE; OTHERWISE;   */
            DO
               nextIndent = nextIndent - 1
            END

            bOneTimeIndention = .false
         END

         CALL Add2Array tmpLine ""
         tmpLine = ""                   /* reset variable for outputting text   */

         bEmptyLinePrinted = .false     /* now an empty line is o.k. again      */
         nextLineBreak = .orx.BreakLinesAt
         ITERATE
      END


      IF tmpLine = "" THEN              /* a new line has to be built           */
      DO
         IF ( LNr - OldLNr > 1 ) & \ bEmptyLinePrinted  THEN       /* at least one empty line was in between in the original */
         DO
            IF bPrintEmptyLines THEN
               CALL Add2Array ""

            bEmptyLinePrinted = .true
/*
call sayerror "--- empty line printed !" "line" pp( lNr ) "oldLnr" pp( oldLnr )
*/
         END
      END

      IF tType = "C" | tType = "CE" THEN     /* comment token in hand        */
      DO
         IF bShowComments THEN          /* comments to be printed ?     */
         DO
            IF \ bCommPrinted THEN      /* if comments not already printed      */
            DO
               bLeave = .false
               beginNr = itemNr
               bCommentBegan = .false
               token_pos = 0
               token_diff = 0

               /* print *all* comments *before* line                       */
               DO tmpInd = itemNr TO arrEntries WHILE anArray[ tmpInd ][ 2 ] <> "N"

                  tmpTType = anArray[ tmpInd ][ 2 ]
                  IF LEFT( tmpTType, 1 ) <> "C" THEN ITERATE

                  tmpToken = anArray[ tmpInd ][ 1 ]

                  IF \ bCommentBegan & tmpTType = "C" THEN   /* check indentation    */
                  DO
                     bCommentBegan = .true
                     token_pos = 0

                     /* determine maximum number of spaces to be truncated from the begin */
                     tmpTTypeX = ""
                     tmpLineX  = 0

                     DO tmpIndX = tmpInd TO arrEntries WHILE tmpTTypeX <> "CE"
                        tmpTokenX = anArray[ tmpIndX ][ 1 ]
                        IF tmpTokenX = "" THEN ITERATE       /* skip empty lines     */
                        tmpTTypeX = anArray[ tmpIndX ][ 2 ]

                        IF tmpLineX = 1 & tmpTTypeX <> "CE" THEN
                        DO
                              tmp_token_pos = VERIFY( tmpTokenX, .orx.WhiteSpace, "N" )
                              IF ( tmp_token_pos - token_diff ) > token_pos THEN
                              DO
                                 token_pos = tmp_token_pos
                              END

                              tmpLineX = tmpLineX + 1
                              ITERATE
                        END

                        IF token_pos > 1 THEN
                        DO
                           token_pos = MIN( token_pos,,
                                       VERIFY( tmpTokenX, .orx.WhiteSpace, "N" ) )
                        END
                        ELSE         /* look for further indentations        */
                           token_pos = VERIFY( tmpTokenX, .orx.WhiteSpace, "N" )

                        tmpLineX  = tmpLineX + 1
                     END
                  END

                  IF bInitialCommentPrinted THEN     /* only if very first comment-block was printed already */
                  DO
                     tmpResult = ""
                     IF bIndent THEN
                        tmpResult = COPIES( sIndent, indent) || sCommIndent

                     tmpResult = tmpResult  ||,
                                 SUBSTR( tmpToken, MAX( 1, MIN( MAX( token_pos, 1 ),,
                                 VERIFY( tmpToken, .orx.WhiteSpace, "N" ) ) ) )
                     CALL Add2Array tmpResult
                  END
                  ELSE
                  DO
                    CALL Add2Array tmpToken
                  END

                  oldLNr = anArray[ tmpInd ][ 4 ]
               END

            END
            bCommPrinted = .true
         END
         ELSE
         DO
/*
            IF tType = "C" | (tType = "CE" & nChar <> "" ) THEN
*/
               oldLNr = lNr
/*
call sayerror "--- a comment" pp(bShowComments ) pp( token ) "line" pp( lNr ) "oldLnr" pp( oldLnr ) "nChar" pp( nChar )
*/
         END

         IF tType = "CE" THEN
         DO
            /* if nothing follows comment immediately, delete oldNChar of previous
               non-comment-token, which contains a "/" */
            IF oldNChar <> "" THEN
               oldNChar = nChar

            bInitialCommentPrinted = .true      /* now indentation is o.k.      */
         END

         ITERATE itemNr
      END


      IF tmpLine = ""       THEN        /* start of a new line  */
      DO
         tmpLine = token

         IF tType = "S" THEN
         DO
            upcToken = TRANSLATE( token )
            IF WORDPOS( upcToken, "IF ELSE DO SELECT WHEN OTHERWISE THEN" ) > 0 THEN
            DO
               nextIndent = indent + 1
               IF   upcToken = "DO" | upcToken = "SELECT" THEN
               DO
                   pendingDo = pendingDo + 1
                   IF WORDPOS( oldFirstWord, "IF ELSE WHEN OTHERWISE" ) > 0 THEN
                   DO
                      nextIndent = indent
                      indent = indent - 1
                   END
               END
               ELSE bOneTimeIndention = .true
            END
            ELSE IF upcToken = "END" THEN       /* decrement indentation                        */
            DO
               indent = indent - 1
               nextIndent = indent
               pendingDo = pendingDo - 1
            END
         END

         oldFirstWord = ""

         /* check for label/procedure/directive !        */
         IF token = ":" THEN                    /* a directive in hand !                        */
         DO
            IF bPrintEmptyLines THEN            /* insert an empty line before directive        */
               CALL Add2Array ""
            bEmptyLinePrinted = .true           /* an empty line was printed                    */

            indent = 0                          /* print directive at column 1                  */
            nextIndent = 1                      /* indent entire rest of directive              */
         END
         ELSE IF anArray[ itemNr + 1][ 1 ] = ":" THEN   /* a label or procedure in hand !       */
         DO
            bProcedure = ( TRANSLATE( anArray[ itemNr + 2][ 1 ] ) = "PROCEDURE" )

            IF \ bProcedure THEN
            DO
               bNewLineFound = .false
               DO tmpItemNr = itemNr + 2 TO arrEntries
                  tmpTType = anArray[ tmpItemNr ][ 2 ]

                  IF \ bNewLineFound THEN
                  DO
                     IF LEFT( tmpTType, 1 ) = "C" THEN ITERATE  /* ignore comments              */
                     ELSE IF tmpTType <> "N" THEN LEAVE         /* no, label is followed by a statement */
                  END

                  /* ok, a new line was found */
                  bNewLineFound = .true
                  IF tmpTType = "N" THEN ITERATE        /* find a non-new line to look for keyword      */

                  bProcedure = ( TRANSLATE( anArray[ tmpItemNr ][ 1 ] ) = "PROCEDURE" )
                  LEAVE
               END
            END

            IF bPrintEmptyLines THEN            /* insert an empty line before procedure        */
               CALL Add2Array ""

            bEmptyLinePrinted = .true           /* an empty line was printed                    */

            IF \ bProcedure THEN                /* a mere label, print it in column 1, but leave indentation in place */
            DO
               nextIndent = indent              /* indentation after label the same             */
               indent   = 0                     /* indentation of label itself at col 1         */
            END
            ELSE                                /* a true procedure was found !                 */
            DO
               indent = 0                       /* print procedure at column 1                  */
               nextIndent = 1                   /* indent entire rest of directive              */
            END
         END

      END
      ELSE IF token = "," | oldNChar = "" THEN tmpLine = tmpLIne token        /* insert a blank before a comma */
      ELSE
      DO                                                        /* abuttal ?            */
         SELECT
            WHEN tType = "O" THEN tmpLine = tmpLine token       /* concatenate operator */
            WHEN tType = "M" THEN
                 DO
                    IF oldToken = "," THEN
                       tmpLine = tmpLine    token    /* comma preceding     */
                    ELSE
                       tmpLine = tmpLine || token    /* abuttal "(", "[" ?   */

                 END
            WHEN tType = "S" THEN                               /* string               */
                 DO
                    IF oldType = "S" THEN                       /* explicit abuttal     */
                       tmpLine = tmpLine "||" token
                    ELSE
                    DO
                       IF oldType = "Q" THEN                    /* quoted string ?      */
                       DO
                          tmpToken = TRANSLATE( token )
                          IF tmpToken = "B" | tmpToken = "X" THEN       /* binary or hex-string ? */
                             tmpLine = tmpLine || token
                          ELSE                                  /* explicit abuttal     */
                          DO
                             tmpLine = tmpLine "||" token
                          END
                       END
                       ELSE                                     /* e.g. an operator ... */
                           tmpLine = tmpLine token
                    END
                 END
            WHEN tType = "Q" THEN
                 DO
                    IF oldType = "S" | oldType = "Q" THEN
                       tmpLine = tmpLine "||" token             /* explicit abuttal     */
                    ELSE
                       tmpLine = tmpLine token
                 END

            OTHERWISE NOP
         END
      END

      /* break up line, because it becomes too long otherwise                           */
      IF WORDPOS( anArray[ itemNr + 1 ][ 2 ], "N C CE" ) = 0 THEN
      DO
         IF LENGTH( tmpLine ) >= nextLineBreak THEN
         DO
            IF bIndent THEN
               tmpLine = tmpLine "," .orx.CRLF || COPIES( sIndent, indent + 1 )  /* yup, and indent next line too        */
             ELSE
               tmpLine = tmpLine "," .orx.CRLF  /* yup, and indent next line too        */


            nextLineBreak = LENGTH( tmpLine ) + .orx.BreakLinesAt + 4 +,
                            LENGTH( COPIES( sIndent, indent + 1 ) ) + LENGTH( .orx.CRLF )
         END
      END

      oldToken = token                  /* remember old token                                   */
      oldType  = tType                  /* remember token type                                  */
      oldNChar = nChar                  /* remember character immediately following token       */
      oldLNr   = MAX( oldLnr, lNr )     /* remember line Nr.                                    */
   END

   IF tmpLine <> "" THEN
      CALL Add2Array tmpLine ""         /* print what's left                                    */

   RETURN OutArray                      /* return array with edited code                        */

Add2Array :                             /* add edited line to array                             */
   IF ARG( 1 ) = "" THEN                /* possibly string of blanks, if so, ignore them        */
      OutArray[ OutIndex ] = ""
   ELSE
      OutArray[ OutIndex ] = ARG( 1 )

   OutIndex = OutIndex + 1
   RETURN






/***********************************************************************************/
/* create a nice cross-reference from token-list, showing all symbols and quotes   */
:: routine Xref                 PUBLIC
   USE ARG anArray

   tmpLine = ""
   maxColsPerLine = 90                  /* chars/line                           */
   arrEntries = anArray ~ items
   tmpDir = .directory ~ new
   maxLength = 0
   lineLength = LENGTH( arrEntries )

   DO itemNr = 1 TO arrEntries
      tType = anArray[ itemNr ][ 2 ]
      IF  tType <> "S" & tType <> "Q" THEN ITERATE      /* just process symbols */

      token  = anArray[ itemNr ][ 1 ]
      lineNr = anArray[ itemNr ][ 4 ]

      IF tType = "Q" THEN                               /* make sure that " is used as quote    */
      DO
         IF LEFT( token, 1 ) <> '"' THEN
         DO
            token = strip_quotes( token )               /* remove ' as quote    */
            token = strip_quotes( token, .false )       /* use " as quote       */
         END

         IF itemNr < arrEntries THEN                    /* determine if binary or hex string    */
         DO
            nextTType = anArray[ itemNr + 1 ][ 2 ]
            IF nextTtype = "S" THEN                     /* a symbol, a "B" or an "X" ?          */
            DO
               nextToken = anArray[ itemNr + 1 ][ 1 ]
               IF POS( TRANSLATE( nextToken ), "BX" ) > 0 THEN /* a binary or hex string found!*/
               DO
                  itemNr = itemNr + 1                   /* increase control variable, to skip this used token */
                  token = token || nextToken
               END
            END
         END
      END

      IF tmpDir ~ hasentry( token ) THEN
      DO
         entry = tmpDir ~ entry( token )                /* get array object                             */
         entry[ 1 ] = token                             /* use this Upper/Lowercase version of symbol   */
         entry[ 2 ] ~ insert( lineNr )                  /* insert line number into list                 */
      END
      ELSE
      DO
         tmpList = .list ~ new
         tmpDir ~ setentry( token, .array ~ of( token, tmpList ~~ insert( lineNr ) ) )
         maxLength = MAX( maxLength, LENGTH( token ) )
      END
   END

   maxLength = MIN( maxLength, 29 )                     /* don't exaggerate with width                  */
   sortedArr = sort( tmpDir )

   maxColsPerLine = maxColsPerLine - lineLength - 1

   DO item OVER sortedArr
      entry = tmpDir ~ entry( item )
      tmpLine = LEFT( entry[ 1 ], MAX( maxLength, LENGTH( entry[ 1 ] ) ) )
      oldLine = 0

      DO lineNr OVER entry[ 2 ]                         /* iterate over line numbers */
         IF lineNr <> oldLine THEN                      /* don't show duplicate line numbers */
         DO
            tmpLine = tmpLine RIGHT( lineNr, lineLength )
            oldLine = lineNr
         END

         IF LENGTH( tmpLine ) > maxColsPerLine then
         DO
            SAY tmpLine
            tmpLine = LEFT( "", maxLength )
         END
      END
      IF tmpLine <> "" THEN
         SAY tmpLine
   END

   RETURN tmpDir





/***********************************************************************************/
/* create a nice cross-reference from token-list, make a distinction between:

        - directives
        - keywords
        - messages
        - quotes
        - routines
        - variables
        - symbols
*/


:: routine Xref_Diff            PUBLIC
   USE ARG someArray

/*
call sayError "someArray ~ items" pp( someArray ~ items ) "someArray ~ dimension" pp( someArray ~ dimension )
*/

   times = someArray ~ items / someArray ~ dimension
   anArray = .array ~ new

   oldNChar   = ""
   oldToken   = ""
   oldCount1  = 1
   bCommentOpen = .false
   anArray[ oldCount1 ] = .array ~ of( "", "N", "", "" )        /* indicate a new line ! */


   DO i = 1 TO times                    /* loop over array, remove comments !           */
      littleArray = someArray[ i ]      /* get "little" array, containing token-infos   */
      token = littleArray[ 1 ]          /* extract token                                */
      tType = littleArray[ 2 ]          /* extract token type                           */
      nChar = littleArray[ 3 ]          /* extract char immediately following token     */
      lineNr= littleArray[ 4 ]          /* extract line number                          */

      SELECT
         WHEN tType = "C" THEN
              DO
                 bCommentOpen = .true
                 ITERATE                /* ignore comment slices                        */
              END
         WHEN tType = "CE" THEN         /* a true end comment                           */
              DO
                 IF oldNChar <> "" THEN
                 DO
                    oldNChar = nChar
                    anArray[ oldCount1 ][ 3 ] = oldNChar /* assign CE' nChar to previous token   */
                 END

                 bCommentOpen = .false
                 oldLineNr = lineNr                     /* make sure there is no linebreak      */
                 ITERATE
              END

         OTHERWISE
              DO
                 oldCount1 = oldCount1 + 1              /* increment counter for entry into array       */
                 oldNChar  = littleArray[ 3 ]           /* save nChar of token                  */
                 oldLineNr = lineNr                     /* save linenumber of token             */
                 anArray[ oldCount1 ] = littleArray     /* add token-array to new array         */
              END
      END
   END


   LengthLineNr = LENGTH( oldLineNr )
   oldCount1 = anArray ~ items + 1
   anArray[ oldCount1 ] = .array ~ of( "", "N", "", "" )/* indicate end of file                 */


   /* create directories for keeping found keywords     */
   tmpDir.Directives = .directory ~ new

   tmpDir.DefinedRoutines = .directory ~ new
   tmpDir.DefinedClasses  = .directory ~ new
   tmpDir.DefinedMethods  = .directory ~ new
   tmpDir.DefinedRequires = .directory ~ new

   tmpDir.Keywords   = .directory ~ new
   tmpDir.Labels     = .directory ~ new
   tmpDir.Messages   = .directory ~ new
   tmpDir.Quotes     = .directory ~ new
   tmpDir.calledRoutines   = .directory ~ new
   tmpDir.Symbols    = .directory ~ new
   tmpDir.Variables  = .directory ~ new

/*
   sequence = "DIRECTIVES KEYWORDS LABELS ROUTINES MESSAGES QUOTES SYMBOLS VARIABLES"  /* "index" for tmpDir. */
*/

   sequence = .array ~ of( .array ~ of( "DIRECTIVES", "DIRECTIVES and symbols used therein" ) ,,
                           .array ~ of( "DEFINEDREQUIRES", "defined REQUIRE(s)"                  ) ,,
                           .array ~ of( "DEFINEDROUTINES", "defined ROUTINE(s)"                  ) ,,
                           .array ~ of( "DEFINEDCLASSES" , "defined/used CLASS(es)"              ) ,,
                           .array ~ of( "DEFINEDMETHODS" , "defined METHOD(s)"                   ) ,,
                           .array ~ of( "LABELS"    , "PROCEDURES/LABELS defined" ) ,,
                           .array ~ of( "CALLEDROUTINES"  , "PROCEDURES/FUNCTIONS/ROUTINES called" ) ,,
                           .array ~ of( "MESSAGES"  , "MESSAGES being sent" ) ,,
                           .array ~ of( "KEYWORDS"  , "KEYWORDS used" ) ,,
                           .array ~ of( "QUOTES"    , "QUOTED strings (outside of directives/keywords)" ) ,,
                           .array ~ of( "SYMBOLS"   , "SYMBOLS (outside of directives)" ) ,,
                           .array ~ of( "VARIABLES" , "VARIABLES" ) ,
                         )

   arrEntries = anArray ~ items / anArray ~ dimension
   lastLineNr = anArray[ arrEntries ][ 4 ]
   tmpLine = ""
   maxColsPerLine = 90                  /* chars/line                           */
   arrEntries = anArray ~ items

   maxLength = 0                        /* maximum length of tokens             */

   oldNChar  = ""
   oldTType  = ""
   oldToken  = ""
   oldLineNr = anArray[ 1 ][ 4 ]        /* first line number                    */
   bNewLine  = .true                    /* indicate a new line                  */

   bKeyword  = .false             /* a keyword is active                  */
   bRoutine  = .false             /* name of a function/procedure expected*/
   actKeyword= ""                 /* keyword-leadin                       */
   actDirective = ""              /* directive-type                       */

   DO itemNr = 1 TO arrEntries
      CALL set_variables

      SELECT
         WHEN tType = "N" THEN
              DO
                 bNewLine = .true       /* a true new line came up              */
                 bExpression = .false           /* expecting an expression              */
                 bKeyword  = .false             /* a keyword is active                  */
                 bRoutine  = .false             /* name of a function/procedure expected*/
                 actKeyword= ""                 /* keyword-leadin                       */
                 actDirective = ""              /* directive-type                       */
              END

         WHEN tType = "S" THEN          /* a symbol                             */
              DO
/*
call sayerror "tType = 'S', lineNr" pp( lineNr ) "token" pp( token ) "ttype" pp( ttype )
*/
                 IF bNewLine THEN       /* check whether a keyword in hand !    */
                 DO
                    tmpLeftNextToken = LEFT( nextToken, 1)
                    bKeyword  = ( nextToken <> "=" & tmpLeftNextToken <> "~" )       /* keyword only, if no assignment !     */
                    IF bKeyword THEN                                    /* check for keyword            */
                    DO
                       actKeyword = .orx.keywords ~ entry( token )      /* attempt to get keyword       */
                       bKeyword = ( actKeyword <> .nil )                /* truly a keyword ?            */
                    END

                    IF bKeyword THEN                                    /* keyword in hand              */
                    DO
                       maxLength = put2Dir( tmpDir.Keywords, actKeyword || "*", lineNr, maxLength )
                       CALL analyzeKeyword
                    END
                    ELSE                                                /* variable/symbol in hand      */
                    DO
/**/
                       /* message to symbol SELF or SUPER ? */
                       IF tmpLeftNextToken = "~" THEN
                       DO
                          IF TRANSLATE( token ) = "SELF" | TRANSLATE( token ) = "SUPER" THEN
                          DO
                             maxLength = put2Dir( tmpDir.Symbols, token, lineNr, maxLength )
                             itemNr = itemNr + 1
                             CALL set_variables
                          END
                       END

                       CALL determine_label                             /* analyze rest of expression */
                    END
                 END
                 ELSE                   /* in the middle of the line            */
                 DO
                    call sayerror "ERROR in the middle of the line, ttype" pp( ttype ) ,
                                  "token" pp( token ) "lineNr" pp( lineNr )
                 END
              END

         WHEN tType = "Q"  THEN         /* a quoted expression                  */
              DO
                 CALL determine_label
              END

/*
         WHEN tType = "O"  THEN         /* an operator                          */
              DO
                 IF ( token = "~" | token = "~~" ) THEN         /* expecting a message  */
                 DO
                    maxLength = put2Dir( tmpDir.Messages  , nextToken, lineNr, maxLength )
                    itemNr = itemNr + 1                         /* get token after nextToekn */
                 END

                 bNewLine = .false
                 ITERATE
              END
*/

         WHEN tType = "M"  THEN         /* miscellaneous, i.e. "," ";" "(" etc. */
              DO
                 bDirective = ( token || nextToken = "::" )     /* expecting a directive                */
                 IF bDirective THEN
                    CALL analyzeDirective
                 ELSE
                    CALL analyzeExpression
              END
         OTHERWISE
              DO
call sayerror "ERROR, never should get here !!, lineNr" pp( lineNr )
call sayerror "    tokenType" pp( tType ) "token" pp( token ) "nTT" pp( nextTType ) "nextToken" pp( nextToken )
              END
      END
   END


   maxLength      = MIN( maxLength, 29 )                /* don't exaggerate with width                  */
   lineLength     = MAX( 4, LengthLineNr )
   maxColsPerLine = maxColsPerLine - lineLength - 1

   DO  i = 1 TO sequence ~ items
      tmpDir = tmpDir.[ sequence[ i ][ 1 ] ]
      IF tmpDir ~ items = 0 THEN ITERATE                /* skip empty directories                       */
      title  = sequence[ i ][ 2 ]
      CALL print_nice tmpDir, title
   END


   RETURN tmpDir.                                       /* return stem                                  */


PRINT_NICE : PROCEDURE EXPOSE maxLength lineLength maxColsPerLine
   USE ARG tmpDir, title

   sortedArr = sort( tmpDir )                           /* sort directory alphabetically        */

   SAY CENTER( " " || pp( title ) || " ", 79, "=" )

   DO item OVER sortedArr
      entry = tmpDir ~ entry( item )
      tmpLine = LEFT( entry[ 1 ], MAX( maxLength, LENGTH( entry[ 1 ] ) ) )
      oldLine = 0

/*
IF tmpLine = "IF*" | tmpLine = "THEN*" | tmpLine = "ELSE*" THEN DO; bbb = .true; CALL sayerror "tmpLine" pp( tmpLine ); END
                  ELSE bbb = .false
*/

      DO lineNr OVER entry[ 2 ]                         /* iterate over line numbers            */

/*
if bbb then call sayerror "    " pp( lineNr )
*/

         IF lineNr <> oldLine THEN                      /* don't show duplicate line numbers    */
         DO
            tmpLine = tmpLine RIGHT( lineNr, lineLength )
            oldLine = lineNr
         END

         IF LENGTH( tmpLine ) > maxColsPerLine THEN
         DO
            SAY tmpLine
            tmpLine = LEFT( "", maxLength )
         END
      END

      IF tmpLine <> "" THEN
         SAY tmpLine
   END
   RETURN




/* a label is a symbol or a string      */
determine_label :
   bLabel = .false
   IF nextToken = ":" THEN                          /* a label, or a message with selector? */
   DO
      DO tmpLoop = 2 TO 3
         IF tType = "N" THEN LEAVE
         IF tmpLoop = 3 THEN
            bLabel = bLabel | ( LEFT( anArray[ itemNr + 3 ][ 1 ], 1) <> "~" )
      END
   END

   IF bLabel THEN
   DO
      maxLength = put2Dir( tmpDir.Labels, token, lineNr, maxLength )
      bNewLine = .true                              /* indicate a new line, so look for keyword "PROCEDURE" or "EXPOSE" */
      itemNr = itemNr + 1                           /* skip next token              */
   END
   ELSE                                             /* something else               */
   DO
      CALL analyzeExpression
   END

   RETURN






/* set variables        */
set_variables :

   token  = anArray[ itemNr ][ 1 ]
   tType  = anArray[ itemNr ][ 2 ]
   nChar  = anArray[ itemNr ][ 3 ]
   lineNr = anArray[ itemNr ][ 4 ]

   IF itemNr > 1 THEN                /* get old token type                   */
   DO
      oldTType = anArray[ itemNr - 1 ][ 2 ]
   END

   IF itemNr <> arrEntries THEN
   DO
      tmpItemNr = itemNr + 1
      nextToken = anArray[ tmpItemNr ][ 1 ]  /* get next token               */
      nextTType = anArray[ tmpItemNr ][ 2 ]  /* get next token type          */
      nextLineNr = anArray[ tmpItemNr ][ 4 ] /* get next line nr             */
   END
   ELSE
   DO
      nextToken = ""
      nextTType = "N"
   END
   RETURN



/* analyze keyword, keyword already saved to directory  */
analyzeKeyword :

   upcKeyword = TRANSLATE( token )
   itemNr = itemNr + 1                          /* position on token, immediately following keyword     */
   CALL set_variables                           /* set variables                                        */

   SELECT
      WHEN  upcKeyword = "ADDRESS" THEN
            DO
               tmpValue = .orx.keywords ~ entry( upcKeyword token )  /* attempt to get keyword       */
               IF tmpValue <> .nil THEN
               DO
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
               END
               ELSE                             /* a symbol, e.g. CMD etc.                              */
               DO
                  CALL analyzeExpression , "SYMBOL"
               END

               itemNr = itemNr + 1           /* position on rest             */

               CALL analyzeExpression
            END

      WHEN  upcKeyword = "ARG" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "CALL" THEN
            DO
               upcToken = TRANSLATE( token )    /* ON | OFF                                     */

               IF upcToken = "ON" | upcToken = "OFF" THEN
               DO
                  /* save keyword ON | OFF */
                  tmpValue = .orx.keywords ~ entry( upcKeyword token )  /* attempt to get keyword  */
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )

                  itemNr = itemNr + 1           /* get next token, e.g. ANY, USER, etc.         */
                  CALL Set_Variables
                  tmpValue = .orx.keywords ~ entry( upcKeyword token )  /* attempt to get keyword       */
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )

                  IF TRANSLATE( token ) = "USER" THEN
                  DO
                     itemNr = itemNr + 1                /* expression after "USER"              */
                     CALL analyzeExpression  "NAME", "SYMBOL"   /* stop at "NAME"                       */
                  END

                  IF upcToken = "ON" THEN               /* look for "NAME"                      */
                  DO
                     IF TRANSLATE( nextToken ) = "NAME" THEN    /* save keyword evaluate expr.  */
                     DO
                        itemNr = itemNr + 1             /* position on "NAME"                   */
                        CALL Set_Variables
                        tmpValue = .orx.keywords ~ entry( upcKeyword token )    /* attempt to get keyword       */
                        maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
                        itemNr = itemNr + 1
                        CALL analyzeExpression "", "SYMBOL"     /* a symbol                     */
                     END
                  END
               END
               ELSE                             /* plain call or "CALL (var)( )"                */
               DO
                  IF token = "(" THEN           /* o.k. "CALL (var) ..." in hand                */
                  DO
                     tmpCounter = 1
                     newToken = token

                     DO UNTIL tmpCounter = 0
                        itemNr = itemNr + 1
                        CALL Set_Variables          /* set variables                                */

                        IF      token = "(" THEN tmpCounter = tmpCounter + 1
                        ELSE IF token = ")" THEN tmpCounter = tmpCounter - 1

                        IF tType = "Q" THEN                             /* a quote in hand ?            */
                           CALL DetermineQuote

                        newToken = newToken token
                     END

                     maxLength = put2Dir( tmpDir.calledRoutines  , newToken , lineNr, maxLength )
                  END
                  ELSE                          /* plain call                                   */
                  DO
                     IF tType = "Q" THEN                                /* a quote in hand ?            */
                        CALL DetermineQuote

                     tmpValue = .orx.functions  ~ entry( token )        /* look for upper/lowercase     */

                     IF tmpValue = .nil THEN tmpValue = token
                                        ELSE tmpValue = tmpValue || "*"

                     maxLength = put2Dir( tmpDir.calledRoutines  , tmpValue    , lineNr, maxLength )
                  END

                  itemNr = itemNr + 1
                  CALL analyzeExpression
               END

            END

      WHEN  upcKeyword = "DO" THEN
            DO
               IF tType <> "N" THEN
               DO
                  upcToken = TRANSLATE( token )
                  IF nextToken = "=" THEN               /* control variable set !       */
                  DO
                     /* starts with an expression setting control variable      */
                     CALL analyzeExpression  "TO BY FOR OVER WHILE UNTIL"
                     IF tType <> "N" THEN
                     DO
                        itemNr = itemNr + 1                                /* position on next token       */
                        CALL Set_Variables
                     END

                     DO WHILE tType <> "N"                              /* as long as no end of line    */
                        upcToken = TRANSLATE( token )

                        tmpValue = .orx.keywords  ~ entry( upcKeyword token )   /* look for upper/lowercase     */
                        maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )

                        itemNr = itemNr + 1                             /* position on expression       */
                        IF upcToken = "WHILE" | upcToken = "UNTIL" THEN /* consume everything           */
                           CALL analyzeExpression
                        ELSE
                           CALL analyzeExpression  "TO BY FOR OVER WHILE UNTIL"

                        IF tType <> "N" THEN
                        DO
                           itemNr = itemNr + 1                          /* position on next token       */
                           CALL Set_Variables
                        END
                     END
                  END
                  ELSE
                  DO
                     tmpValue = .orx.keywords  ~ entry( upcKeyword token )   /* look for upper/lowercase     */
                     IF tmpValue <> .nil THEN                   /* FOREVER/WHILE/UNTIL ? */
                     DO
                        maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
                        itemNr = itemNr + 1
                     END

                     IF upcKeyWord <> "FOREVER" THEN
                        CALL analyzeExpression                          /* expression   */
                  END
               END
            END

      WHEN  upcKeyword = "DROP" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "ELSE" THEN
            DO
               itemNr = itemNr - 1      /* put present token back               */
            END

      WHEN  upcKeyword = "END"  THEN
            DO
               itemNr = itemNr - 1      /* put present token back               */
            END

      WHEN  upcKeyword = "EXIT" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "EXPOSE" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "FORWARD" THEN
            DO
               DO WHILE tType <> "N"                            /* as long as no end of line    */
                  upcToken = TRANSLATE( token )                 /* get next token               */

                  tmpValue = .orx.keywords  ~ entry( upcKeyword token )/* look for upper/lowercase     */
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )

                  IF upcToken <> "CONTINUE" THEN                /* followed by an expression    */
                  DO
                     itemNr = itemNr + 1                        /* position on expression       */
                     CALL analyzeExpression  "CONTINUE ARGUMENTS ARRAY MESSAGE CLASS TO"
                  END

                  itemNr = itemNr + 1                           /* position on next token       */
                  CALL Set_Variables
               END
               itemNr = itemNr - 1                              /* let main loop determine end  */
            END

      WHEN  upcKeyword = "IF" THEN
            DO
               CALL analyzeExpression  "THEN"                   /* stop at "THEN"               */
               tmpValue = .orx.keywords ~ entry( upcKeyword nextToken )  /* attempt to get keyword       */
/*
call sayerror "in IF -- upcKeyword" pp( upcKeyword ) "nextToken" pp( nextToken ) "tmpValue" pp( tmpValue ) "linenr" pp( nextLineNr )
*/
               IF tmpValue <> .nil THEN
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", nextLineNr, maxLength )
               itemNr = itemNr + 1
            END

      WHEN  upcKeyword = "GUARD" THEN
            DO
               tmpValue = .orx.keywords ~ entry( upcKeyword token )  /* attempt to get keyword       */
               maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
               itemNr = itemNr + 1

               IF TRANSLATE( nextToken ) = "WHEN" THEN
               DO
                  tmpValue = .orx.keywords ~ entry( upcKeyword nextToken )  /* attempt to get keyword       */
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
                  itemNr = itemNr + 1
               END
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "INTERPRET" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "ITERATE" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "LEAVE" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "NOP" THEN
            DO
               itemNr = itemNr - 1      /* put present token back               */
            END

      WHEN  upcKeyword = "NUMERIC" THEN
            DO
               tmpValue = .orx.keywords ~ entry( upcKeyword Token ) /* attempt to get keyword       */
               maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
               itemNr = itemNr + 1

               IF TRANSLATE( token ) = "FORM" THEN
               DO
                  CALL Set_Variables                                     /* get next "record"            */
                  upcToken = TRANSLATE( token )
                  tmpValue = .orx.keywords ~ entry( upcKeyword Token )  /* attempt to get keyword       */
                  IF tmpValue <> .nil THEN
                  DO
                     maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
                     itemNr = itemNr + 1
                  END
               END
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "OPTIONS" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "OTHERWISE" THEN
            DO
               itemNr = itemNr - 1      /* put present token back               */
            END

      WHEN  upcKeyword = "PARSE" THEN
            DO
               upcToken = TRANSLATE( token )
               DO 2
                  IF upcToken = "UPPER" | upcToken = "LOWER" | upcToken = "CASELESS" THEN
                  DO
                     tmpValue = .orx.keywords ~ entry( upcKeyword Token )  /* attempt to get keyword       */
                     maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )

                     itemNr = itemNr + 1        /* get next token                       */
                     CALL Set_Variables
                     upcToken = TRANSLATE( token )
                  END
                  ELSE                          /* no UPPER | LOWER | CASELESS found    */
                     LEAVE
               END

               /* save ARG, LINEIN, etc.        */
               tmpValue = .orx.keywords ~ entry( upcKeyword Token )  /* attempt to get keyword       */
               maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )

               IF upcToken = "VALUE" THEN
               DO
                  itemNr = itemNr + 1
                  CALL Set_Variables
                  CALL analyzeExpression  "WITH"        /* stop at WITH                 */

                  itemNr = itemNr + 1
                  CALL Set_Variables
                  tmpValue = .orx.keywords ~ entry( upcKeyword Token )  /* attempt to get keyword       */
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
               END

               itemNr = itemNr + 1
               CALL analyzeExpression           /* get tokenlist                        */
            END

      WHEN  upcKeyword = "PROCEDURE" THEN
            DO
               itemNr = itemNr - 1      /* put present token back               */
/*
call sayerror
call sayerror "upcKeyword = 'PROCEDURE', lineNr" pp( lineNr ) "token" pp( token ) "nextToken" pp( nextToken )
call sayerror "       tmpValue" pp( tmpValue )
call set_variables
call sayerror "upcKeyword = 'PROCEDURE', lineNr" pp( lineNr ) "token" pp( token ) "nextToken" pp( nextToken ) "(after setting)"
do rgf = itemNr to 1 by -1
   call sayerror "itemNr" pp( rgf ),
                 "lineNr" pp( anArray[ rgf ][ 4 ] ),
                 "token" pp( anArray[ rgf ][ 1 ] ),
                 "tType" pp( anArray[ rgf ][ 2 ] )

   if TRANSLATE( anArray[ rgf ][ 1 ] ) = "PROCEDURE" then leave
end /* do */

*/
            END

      WHEN  upcKeyword = "PULL" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "PUSH" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "QUEUE" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "RAISE" THEN
            DO
               tmpValue = .orx.keywords  ~ entry( upcKeyword token )   /* look for upper/lowercase     */
               maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
               upcToken = TRANSLATE( token )

               /* conditions followed by expression             */
               IF WORDPOS( upcToken, "ERROR FAILURE SYNTAX USER" ) > 0 THEN
               DO
                  itemNr = itemNr + 1
                  IF upcToken = "USER" THEN                     /* usercondition is a symbol only       */
                     CALL analyzeExpression "ADDITIONAL ARRAY DESCRIPTION EXIT RETURN", "SYMBOL"
                  ELSE
                     CALL analyzeExpression "ADDITIONAL ARRAY DESCRIPTION EXIT RETURN"

                  itemNr = itemNr + 1
                  CALL set_Variables
               END

               /* analyze options                               */
               DO WHILE tType <> "N"                            /* as long as no end of line    */
                  upcToken = TRANSLATE( token )                 /* get next token               */

                  tmpValue = .orx.keywords  ~ entry( upcKeyword token )   /* look for upper/lowercase     */
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )

                  itemNr = itemNr + 1                           /* position on expression       */
                  CALL analyzeExpression  "ADDITIONAL ARRAY DESCRIPTION EXIT RETURN"

                  itemNr = itemNr + 1                           /* position on next token       */
                  CALL Set_Variables
               END
               itemNr = itemNr - 1                              /* let main loop determine end  */
            END

      WHEN  upcKeyword = "REPLY" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "RETURN" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "SAY" THEN
            DO
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "SELECT" THEN
            DO
               itemNr = itemNr - 1      /* put present token back               */
            END


      WHEN  upcKeyword = "SIGNAL" THEN
            DO
               upcToken = TRANSLATE( token )    /* VALUE | ON | OFF                             */

               IF WORDPOS( upcToken, "ON OFF VALUE" ) > 0 THEN
               DO
                  /* save keyword ON | OFF | VALUE*/
                  tmpValue = .orx.keywords ~ entry( upcKeyword token )  /* attempt to get keyword  */
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )

                  IF upcToken = "VALUE" THEN    /* merely get rest as expression                */
                  DO
                     itemNr = itemNr + 1        /* rest is an expression                        */
                     CALL analyzeExpression
                  END
                  ELSE                          /* ON | OFF                                     */
                  DO
                     itemNr = itemNr + 1           /* get next token, e.g. ANY, USER, etc.         */
                     CALL Set_Variables
                     tmpValue = .orx.keywords ~ entry( upcKeyword token )  /* attempt to get keyword       */
                     maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )

                     IF TRANSLATE( token ) = "USER" THEN
                     DO
                        itemNr = itemNr + 1                /* expression after "USER"              */
                        CALL analyzeExpression  "NAME", "SYMBOL"   /* stop at "NAME"                       */
                     END

                     IF upcToken = "ON" THEN               /* look for "NAME"                      */
                     DO
                        IF TRANSLATE( nextToken ) = "NAME" THEN    /* save keyword evaluate expr.  */
                        DO
                           itemNr = itemNr + 1             /* position on "NAME"                   */
                           CALL Set_Variables
                           tmpValue = .orx.keywords ~ entry( upcKeyword token )    /* attempt to get keyword       */
                           maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
                           itemNr = itemNr + 1
                           CALL analyzeExpression
                        END
                     END
                  END
               END
               ELSE                             /* get labelname                */
               DO
                  CALL analyzeExpression , "SYMBOL"
               END
            END


      WHEN  upcKeyword = "THEN" THEN
            DO
               itemNr = itemNr - 1      /* put present token back               */
            END


      WHEN  upcKeyword = "TRACE" THEN
            DO
               upcToken = TRANSLATE( STRIP( token, "L", "?" ) ) /* get rid of ?                 */
               bExpr = .false

               bExpr = DATATYPE( upcToken, "W" )/* a whole number ?     */

               IF \ bExpr & ( upcToken = "VALUE" ) THEN /* keyword part, save, evaluate expr.   */
               DO
                  bExpr = .true                 /* evaluate expr                                */
                  tmpValue = .orx.keywords ~ entry( upcKeyword token )  /* attempt to get keyword       */
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
                  itemNr = itemNr + 1
               END
               ELSE
               DO
                  tmpAbbrWords = .array ~ of( "ALL", "COMMANDS", "ERROR", "FAILURE", "INTERMEDIATES",,
                                              "LABELS", "NORMAL", "OFF", "RESULTS" )

                  bExpr = .true                         /* default to expression                */
                  DO tmpItem OVER tmpAbbrWords
                     IF ABBREV( tmpItem , upcToken ) THEN
                     DO
                        bExpr = .false                  /* not an expression, an abbreviation ! */
                        tmpValue = .orx.keywords ~ entry( upcKeyword tmpItem )  /* attempt to get keyword       */
                        maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
                        itemNr = itemNr + 1
                     END
                  END
               END

               IF bExpr THEN                            /* analyze expression                   */
                  CALL analyzeExpression
            END

      WHEN  upcKeyword = "USE" THEN
            DO
               tmpValue = .orx.keywords ~ entry( upcKeyword token )  /* attempt to get keyword       */
               IF tmpValue <> .nil THEN
               DO
                  maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
                  itemNr = itemNr + 1                                /* position on rest             */
               END
               CALL analyzeExpression
            END

      WHEN  upcKeyword = "WHEN" THEN
            DO
               CALL analyzeExpression  "THEN"          /* stop at THEN */
               tmpValue = .orx.keywords ~ entry( upcKeyword nextToken )  /* attempt to get keyword       */
               maxLength = put2Dir( tmpDir.Keywords, tmpValue || "*", lineNr, maxLength )
               itemNr = itemNr + 1
            END

      OTHERWISE
do
call sayerror "ERROR ! analyzeKeywords:OTHERWISE(), unhandled token:" pp( upcKeyword )
call sayerror "        token" pp( token ) pp( ttype ) "next" pp( nextToken ) pp( nextTType )
      NOP
end
   END
   bNewLine = .true

   RETURN



/* analyze expression, "itemNr" has to point right *at* first expression-token   */
analyzeExpression :
   USE ARG exitTokens, tmpDataType          /* indicates end of expression, e.g. THEN following an IF-statement !       */

   IF \ VAR( "exitTokens" ) THEN exitTokens = ""

/*
call sayerror
call sayerror "-----------------------"
call sayerror "analyzeExpression(), exitTokens" pp( exitTokens ) "token" pp( token ) "itemNr" pp( itemNr )
*/

   tmpOpen = 0                                  /* nr. of open parenthesis              */
   bJustOpenedParenthesis = .false
   DO itemNr = itemNr TO arrEntries
      CALL set_variables
      SELECT
         WHEN tType = "N" THEN                  /* new line, end                        */
              DO
                 itemNr = itemNr -1
                 RETURN
              END

         WHEN tType = "M" THEN                  /* just take care of opening/closing parenthesis        */
              DO
                 IF      token = "(" THEN
                 DO
                    tmpOpen = tmpOpen + 1
                    bJustOpenedParenthesis = .true
                    ITERATE
                 END
                 ELSE IF token = ")" THEN tmpOpen = tmpOpen - 1
                 ELSE IF token = ":" THEN                       /* expecting a qualifier, i.e. a symbol */
                 DO
/**/
call sayerror "  qualifier found for" pp( nextToken ) "L#" pp( nextLineNr )
                    maxLength = put2Dir( tmpDir.Symbols, nextToken, nextLineNr, maxLength )
                    itemNr = itemNr + 1                         /* upcoming token already handled, so no function call */
                 END
              END

         WHEN tType = "O" THEN                  /* just take care of message operator   */
              DO
                 IF ( token = "~" | token = "~~" ) THEN         /* expecting a message  */
                 DO
                    maxLength = put2Dir( tmpDir.Messages  , nextToken, lineNr, maxLength )
                    itemNr = itemNr + 1                         /* get token after nextToken */
                 END
              END

         WHEN tType = "S" | tType = "Q" THEN    /* handle symbols, determine routines, variables, quotes */
              DO
                 IF tType = "Q" THEN                            /* a quote in hand ?            */
                 DO
                    CALL DetermineQuote
                 END
                 ELSE IF bJustOpenedParenthesis THEN            /* check for SELF or SUPER being a target for a message */
                 DO
                    IF LEFT( nextToken, 1 ) = "~" THEN          /* a message being sent ?       */
                    DO
                       IF TRANSLATE( token ) = "SELF" | TRANSLATE( token ) = "SUPER" THEN
                       DO
                          maxLength = put2Dir( tmpDir.Symbols, token, lineNr, maxLength )
                          ITERATE
                       END

                    END
                 END

                 IF nChar <> "" & nextToken = "(" THEN  /* calling a routine            */
                 DO
                    tmpValue = .orx.functions  ~ entry( token )           /* look for upper/lowercase     */
                    IF tmpValue = .nil THEN tmpValue = token
                                       ELSE tmpValue = tmpValue || "*"
                    maxLength = put2Dir( tmpDir.calledRoutines  , tmpValue    , lineNr, maxLength )
                 END
                 ELSE                           /* normal symbol | quote                */
                 DO
/*
call sayerror "--- landed in analyzeExpression ..., lineNr" pp( lineNr ) "token" pp( token )
*/
                    IF tType = "Q" THEN                 /* quoted value                         */
                       maxLength = put2Dir( tmpDir.Quotes, token, lineNr, maxLength )
                    ELSE IF LEFT( token, 1 ) = "." | tmpDataType = "SYMBOL" THEN /* a symbol starting with a dot         */
                       maxLength = put2Dir( tmpDir.Symbols, token, lineNr, maxLength )
                    ELSE IF DATATYPE( token, "N" ) THEN /* numeric value                        */
                       maxLength = put2Dir( tmpDir.Symbols, token, lineNr, maxLength )
                    ELSE                                /* plain variable                       */
                       maxLength = put2Dir( tmpDir.Variables, token, lineNr, maxLength )
                 END
              END

         OTHERWISE NOP                          /* ignore all other types               */
      END
      bJustOpenedParenthesis = .false

/*
call sayerror "analyzeExpression(), nr. of open parenthesis" pp( tmpOpen )
*/

      IF tmpOpen = 0 THEN            /* check whether expression has to be ended */
      DO                             /* e.g. IF ... THEN, WHEN ... THEN,         */
         IF nextTType = "S" THEN     /*      DO ... TO ... BY ... FOR, etc.      */
         DO
/*
call sayerror "analyzeExpression(), testing whether next token is in exitTokens" pp( nextToken )
*/
            IF WORDPOS( TRANSLATE( nextToken ), exitTokens ) > 0 THEN
            DO
/*
call sayerror "analyzeExpression(), BINGO, premature return !"
call sayerror
*/
               RETURN
            END
         END
      END
   END

   RETURN


/* analyze directives   */
analyzeDirective :

   /* get directive itself     */
   itemNr = itemNr + 2                          /* get token after 2nd ":"              */
   nextToken = anArray[ itemNr ][ 1 ]
   lineNr = anArray[ itemNr ][ 4 ]

   actDirective = .orx.directives ~ entry( nextToken )
   upActDirective = TRANSLATE( actDirective )
   maxLength = put2Dir( tmpDir.Directives, actDirective || "*"   , lineNr, maxLength )

   /* save name for directive                                                           */
   itemNr = itemNr + 1                          /* get name of directory                */
   CALL set_variables                           /* set variables from array             */


   SELECT
      WHEN upActDirective = "REQUIRES" THEN
           DO
              maxLength = put2Dir( tmpDir.DefinedRequires, token , lineNr, maxLength )
           END

      WHEN upActDirective = "ROUTINE" THEN
           DO
              maxLength = put2Dir( tmpDir.DefinedRoutines, token , lineNr, maxLength )

              /* check nextToken        */
              tmpValue = .orx.directives ~ entry( upActDirective nextToken )
              IF tmpValue <> .nil THEN               /* part of directive               */
              DO
                 maxLength = put2Dir( tmpDir.Directives, tmpValue || "*"   , nextLineNr, maxLength )
                 itemNr = itemNr + 1
              END
           END

      WHEN upActDirective = "CLASS" THEN
           DO
              maxLength = put2Dir( tmpDir.DefinedClasses, token , lineNr, maxLength )

              DO itemNr = itemNr + 1 TO arrEntries      /* get rest of definitions, i.e. PUBLIC, if there */
                 CALL set_variables

                 upToken = TRANSLATE( token )           /* put token into uppercase     */

                 tmpValue = .orx.directives ~ entry( actDirective token )
                 IF tmpValue <> .nil THEN         /* part of directive            */
                 DO
                        /* got a directive      */
                    maxLength = put2Dir( tmpDir.Directives, tmpValue || "*"   , lineNr, maxLength )

                    IF upToken = "PUBLIC" THEN
                       ITERATE

                    ELSE IF upToken = "INHERIT" THEN    /* process list of class names S | Q, always last   */
                    DO
                       DO itemNr = itemNr + 1 TO arrEntries
                          CALL set_variables
                          IF tType = "S" | tType = "Q" THEN
                          DO
                             maxLength = put2Dir( tmpDir.DefinedClasses, token , lineNr, maxLength )
                          END
                          ELSE
                             LEAVE
                       END
                       itemNr = itemNr - 1
                       LEAVE
                    END
                    ELSE                                /* next token must be a class !         */
                    DO
/*
                       maxLength = put2Dir( tmpDir.Directives, nextToken , nextLineNr, maxLength )
*/
                       itemNr = itemNr + 1
                       CALL Set_Variables
                       maxLength = put2Dir( tmpDir.DefinedClasses, token , lineNr, maxLength )
/*
                       itemNr = itemNr + 1
*/
                    END
                 END
                 ELSE                                   /* nothing found anymore        */
                 DO
                    itemNr = itemNr - 1
                    LEAVE
                 END
              END
           END

      WHEN upActDirective = "METHOD" THEN
           DO
              maxLength = put2Dir( tmpDir.DefinedMethods, token , lineNr, maxLength )
              DO itemNr = itemNr + 1 TO arrEntries      /* get rest of definitions, i.e. PUBLIC, if there */
                 CALL set_variables

                 tmpValue = .orx.directives ~ entry( actDirective token )
                 IF tmpValue <> .nil THEN         /* part of directive            */
                 DO
                    maxLength = put2Dir( tmpDir.Directives, tmpValue || "*"   , lineNr, maxLength )
                 END
                 ELSE                                   /* nothing found anymore        */
                 DO
                    itemNr = itemNr - 1
                    LEAVE
                 END
              END
           END

      OTHERWISE
           CALL sayerror "analyzeDirective() - unknown directive" pp( actDirective )
   END

   RETURN



/* determine whether quote is in effect a binary or hexadecimal string          */
DetermineQuote :
    IF LEFT( token, 1 ) <> '"' THEN             /* use only " as quotes */
    DO
       token = strip_quotes( token )            /* remove ' as quote    */
       token = strip_quotes( token, .false )    /* use " as quote       */
    END

    IF POS( TRANSLATE( nextToken ), "BX" ) > 0 THEN /* a binary or hex string found!    */
    DO
       tmpToken = token || nextToken
       itemNr = itemNr + 1                      /* increase control variable, to skip this used token */
       CALL Set_Variables

       token  = tmpToken                        /* reset token itself   */
       tType  = "Q"                             /* reset token Type to quotes   */
    END
    RETURN





/* put token into appropriate directory, return present maximuml length of token */
PUT2DIR : PROCEDURE
   USE ARG tmpDir, token, lineNr, maxLength

   IF tmpDir ~ hasentry( token ) THEN
   DO
      entry = tmpDir ~ entry( token )                /* get array object                             */
      entry[ 1 ] = token                             /* use this Upper/Lowercase version of symbol   */
      entry[ 2 ] ~ insert( lineNr )                  /* insert line number into list                 */
   END
   ELSE
   DO
      tmpList = .list ~ new
      tmpDir ~ setentry( token, .array ~ of( token, tmpList ~~ insert( lineNr ) ) )
      maxLength = MAX( maxLength, LENGTH( token ) )
   END
   RETURN maxLength


/* dump directories     */
DumpDir : PROCEDURE
   USE ARG tmpDir, maxLength, lineLength, maxColsPerLine

   sortedArr = sort( tmpDir )                           /* sort directory                       */
   DO item OVER sortedArr
      entry = tmpDir ~ entry( item )
      tmpLine = LEFT( entry[ 1 ], MAX( maxLength, LENGTH( entry[ 1 ] ) ) )
      oldLine = 0
      IF entra[ 2 ] ~ items > 0 THEN                    /* dump entries                         */
      DO
         DO lineNr OVER entry[ 2 ]                      /* iterate over line numbers */
            IF lineNr <> oldLine THEN                   /* don't show duplicate line numbers */
            DO
               tmpLine = tmpLine RIGHT( lineNr, lineLength )
               oldLine = lineNr
            END

            IF LENGTH( tmpLine ) > maxColsPerLine then
            DO
               SAY tmpLine
               tmpLine = LEFT( "", maxLength )
            END
         END

         IF tmpLine <> "" THEN
            SAY tmpLine
      END
      ELSE
         SAY CENTER( "--- no entries available ---", maxColsPerLine )

   END
   RETURN






/******************************************************************************/
/*                                                                            */
/* name:    token_in_string( line, token_position, line# )                    */
/*                                                                            */
/* purpose: checks whether a specific position falls within quotes            */
/*                                                                            */
/* returns: .true, if token is within quotes, .false else                     */
/*                                                                            */
/* remarks: ---                                                               */
/*                                                                            */
/* created: rgf, 95-10-20                                                     */
/* check whether a given stringpos falls into quotes, i.e. part of a string */

:: ROUTINE token_in_string       PUBLIC
  USE ARG line, stringpos, line?

  SIGNAL ON SYNTAX
  IF \VAR( "line?" ) THEN line? = "line # not available ..."

  quotes = '"' || "'"
  deli_pos = VERIFY( line, quotes, "M" )

  IF deli_pos > stringpos THEN RETURN .false

  IF deli_pos = 0 THEN RETURN .false     /* not in string */

  deli      = SUBSTR( line, deli_pos, 1 )                     /* get quote */

  DO FOREVER
     deli_next = VERIFY( line, deli, "Match", deli_pos + 1 )    /* get next quote */

     IF deli_next = 0 THEN      /* error, ' or " not closed on the same line ! */
     DO
        SAY "*** error analyzing line #" pp( line? ) "quote" pp( deli ) "not closed !"
        IF deli = "'" THEN RAISE SYNTAX 6.2     /* unmatched ' */
                      ELSE RAISE SYNTAX 6.3     /* unmatched " */
     END

     IF deli_next > stringpos THEN RETURN .true         /* stringpos within quotes */

     IF SUBSTR( line, deli_next + 1, 1) = deli THEN     /* quote escaped, still in string */
     DO
       deli_pos = deli_next + 2
       ITERATE
     END


     deli_pos = VERIFY( line, quotes, "Match", deli_next + 1 )

     IF deli_pos > stringpos THEN RETURN .true

     IF deli_pos = 0 THEN RETURN .false                 /* stringpos not in string      */

     deli = SUBSTR( line, deli_pos, 1 )                 /* get new delimiter            */
  END

  RETURN


SYNTAX : RAISE PROPAGATE                                /* raise SYNTAX error in caller */
/******************************************************************************/
/******************************************************************************/



/******************************************************************************/
/*                                                                            */
/* name:    get_next_token( line, token_position, line? )                     */
/*                                                                            */
/* purpose: extracts next token, i.e. string, constant symbol, delimiters     */
/*                                                                            */
/* returns: array of ( token, tokentype, next token_position,                 */
/*                     char immediately following )                           */
/*                                                                            */
/*          token_type:                                                       */
/*                                                                            */
/*          S ... Symbol                                                      */
/*          Q ... Quoted String                                               */
/*          O ... Operator                                                    */
/*          M ... Miscellaneous, like ();.:[] etc.                            */
/*          '' ... nothing left (whitespace only)                             */
/*                                                                            */
/* remarks: no comments allowed (would screw up this routine)                 */
/*                                                                            */
/* created: rgf, 96-02-08                                                     */

:: ROUTINE get_next_token        PUBLIC
  USE ARG line, token_pos, line?

  SIGNAL ON SYNTAX
  IF \VAR( "line?" ) THEN line? = "line # not available ..."

  token_pos = VERIFY( line, .orx.WhiteSpace, "N", token_pos )   /* skip whitespace characters */
  IF token_pos = 0 THEN /* no tokens left */
     RETURN .array ~ of( "", "", 0, "" )

  quotes = '"' || "'"
  deli_pos = POS( SUBSTR( line, token_pos, 1) , quotes )

  IF deli_pos > 0 THEN                          /* o.k. a string token in hand  */
  DO
     deli      = SUBSTR( line, token_pos, 1 )           /* extract type of deli */
     deli_pos = token_pos
     DO FOREVER
        deli_next = VERIFY( line, deli, "Match", deli_pos + 1 )    /* get next quote */

        IF deli_next = 0 THEN      /* error, ' or " not closed on the same line ! */
        DO
           CALL sayerror "*** error analyzing line #" pp( line? ) "quote" pp( deli ) "not closed !"
           IF deli = "'" THEN RAISE SYNTAX 6.2  /* unmatched ' */
                         ELSE RAISE SYNTAX 6.3  /* unmatched " */
        END

        IF SUBSTR( line, deli_next + 1, 1) = deli THEN  /* quote escaped, still in string       */
        DO
          deli_pos = deli_next + 1
          ITERATE
        END

        /* end of string token found            */

        tmpToken = SUBSTR( line, token_pos, deli_next - token_pos + 1 )
        IF deli_next = LENGTH( line ) THEN
        DO
           token_pos = 0      /* no more tokens left on line ?        */
           nextChar  = ""
        END
        ELSE
        DO
           token_pos = deli_next + 1
           nextChar  = SUBSTR( line, token_pos, 1 )     /* char *immediately* following */
        END
        RETURN .array ~ of( tmpToken , "Q", token_pos, nextChar )
     END
  END


  /* get first non-white-char */
  bIsSymbol = ( VERIFY( SUBSTR( line, token_pos, 1) , .orx.SymbolChars, "M" ) = 1 )
  IF bIsSymbol THEN                             /* o.k. a symbol token in hand  */
  DO
     deli_next = VERIFY( line, .orx.SymbolChars, "N", token_pos + 1 )   /* get next quote       */
     IF deli_next = 0 THEN                      /* o.k. only symbol left                        */
        RETURN .array ~ of( SUBSTR( line, token_pos), "S", 0, "" )

     /* extract symbol and return it    */
     tmpToken = SUBSTR( line, token_pos, deli_next - token_pos )
     IF deli_next - 1 = LENGTH( line ) THEN
     DO
        token_pos = 0      /* no more tokens left on line ?        */
        nextChar  = ""
     END
     ELSE
     DO
        token_pos = deli_next
        nextChar  = SUBSTR( line, token_pos, 1 )     /* char *immediately* following */
     END

     RETURN .array ~ of( tmpToken , "S", token_pos, nextChar )
  END

  bIsOperator = ( VERIFY( SUBSTR( line, token_pos, 1) , .orx.OperatorChars, "M" ) = 1 )
  IF bIsOperator THEN                           /* o.k. a Operator token in hand                */
  DO
     deli_next = VERIFY( line, .orx.OperatorChars, "N", token_pos + 1 ) /* get next non-operator */
     IF deli_next = 0 THEN                      /* o.k. only Operator left                       */
        RETURN .array ~ of( SUBSTR( line, token_pos), "O", 0, "" )

     /* extract Operator and return it    */
     tmpToken = SUBSTR( line, token_pos, deli_next - token_pos )

     IF deli_next - 1 = LENGTH( line ) THEN
     DO
        token_pos = 0      /* no more tokens left on line ?        */
        nextChar  = ""
     END
/*???????????????????????????????????? commmented assignement of ELSE ! */
     ELSE /* token_pos = deli_next */
     DO
        token_pos = deli_next
        nextChar  = SUBSTR( line, token_pos, 1 )     /* char *immediately* following */
     END

     RETURN .array ~ of( tmpToken , "O", token_pos, nextChar )
  END


  /* o.k. now only quotes " or ' ~ ( ) [ ] : ; , or white space left */
  tmpToken = SUBSTR( line, token_pos, 1 )
  IF token_pos = LENGTH( line ) THEN
  DO
     token_pos = 0      /* no more tokens left on line ?        */
     nextChar  = ""
  END
  ELSE
  DO
     token_pos = token_pos + 1
     nextChar  = SUBSTR( line, token_pos, 1 )     /* char *immediately* following */
  END

  RETURN .array ~ of( tmpToken, "M", token_pos, nextChar )



SYNTAX : RAISE PROPAGATE                                /* raise SYNTAX error in caller */
/******************************************************************************/
/******************************************************************************/




/******************************************************************************/
/*                                                                            */
/* name:    make_token_list( line, bWorkOnFile )                              */
/*                                                                            */
/* purpose: extracts Rexx-tokens, i.e. string, constant symbol, delimiters,   */
/*          *and* comments from a line or an entire file (default)            */
/*                                                                            */
/* returns: array of array of ( token, tokentype,                             */
/*                              char immediately following, line# )           */
/*                                                                            */
/*          token_type:                                                       */
/*                                                                            */
/*          C ... Comment                                                     */
/*          CE... Comment End                                                 */
/*          S ... Symbol                                                      */
/*          Q ... Quoted String                                               */
/*          O ... Operator                                                    */
/*          M ... Miscellaneous, like ();.:[] etc.                            */
/*          '' ... nothing left (whitespace only)                             */
/*                                                                            */
/* remarks: if bWorkOnFile is .true, then "line" is in effect file-name       */
/*                                                                            */
/* created: rgf, 96-02                                                        */

:: ROUTINE make_token_list        PUBLIC
  USE ARG line, bWorkOnFile

  SIGNAL ON SYNTAX
  IF \VAR( "bWorkOnFile" ) THEN bWorkOnFile = .true     /* first argument a Rexx line by default */

/*
say "bWorkOnFile" pp( bWorkOnFile )
*/

  IF bWorkOnFile THEN                   /* analyze an entire Rexx file          */
     tmpArr = .stream ~ new( line ) ~~ open( "read" ) ~ arrayin         /* read file's content into array       */
  ELSE
     tmpArr = .array ~ of( line )       /* put line to be analyzed into array   */

  resultArray   = .array ~ new          /* array to contain arrays with tokens  */
  resIndex      = 1                     /* index into result array              */

  totLines = tmpArr ~ items             /* get total lines                      */


  bNewLine      = .false                /* emit a new-line indicator ?          */
  bCommaPending = .false                /* a comma pending ?                    */

  DO lineNr = 1 TO totLines             /* loop over array                      */
/*
say "working on lineNr" pp( lineNr )
*/
     token_pos     = 1                  /* start from the beginning of the line */
     line    = tmpArr[ lineNr ]         /* get next line to analyze             */
     iLineLength = LENGTH( line )       /* get length of line                   */

     IF line = "" THEN ITERATE

     DO workLine = 1 TO 2 BY 0          /* make sure a forever statement, control variable serves as target     */

        token_pos = VERIFY( line, .orx.WhiteSpace, "N", token_pos )   /* skip whitespace characters */

        IF ( bNewLine | token_pos = 0 ) & \ bCommaPending then
        DO
/*
call sayerror "in BOT l#" pp( right( lineNr, 4 ) ) "comma pending" pp( bCommaPending ) pp( tmpToken )
*/

           resultArray[ resIndex ] = .array ~ of( "", "N", "", "" )     /* indicate new line    */
           resIndex = resIndex + 1   /* calculate next entry                         */

           bNewLine = .false
           bCommaPending = .false
        END

        IF token_pos = 0 THEN           /* no tokens left on line               */
           ITERATE lineNr               /* go, get next line                    */


        /* ---> extract quoted text  */
        quotes = '"' || "'"
        deli_pos = POS( SUBSTR( line, token_pos, 1) , quotes )


/**********************************************************************************/
        IF deli_pos > 0 THEN                          /* o.k. a string token in hand  */
        DO
           CALL Pending_Comma
           deli      = SUBSTR( line, token_pos, 1 )     /* extract type of deli */
           deli_pos = token_pos
           DO FOREVER
              deli_next = VERIFY( line, deli, "Match", deli_pos + 1 )    /* get next quote */

              IF deli_next = 0 THEN      /* error, ' or " not closed on the same line ! */
              DO
                 CALL sayerror "*** error analyzing line #" pp( line? ) "quote" pp( deli ) "not closed !"
                 IF deli = "'" THEN RAISE SYNTAX 6.2  /* unmatched ' */
                               ELSE RAISE SYNTAX 6.3  /* unmatched " */
              END

              IF SUBSTR( line, deli_next + 1, 1) = deli THEN  /* quote escaped, still in string       */
              DO
                deli_pos = deli_next + 1
                ITERATE
              END

              /* end of string token found            */

              tmpToken = SUBSTR( line, token_pos, deli_next - token_pos + 1 )
              IF deli_next = iLineLength    THEN        /* at end of line, nor chars following  */
              DO
                 token_pos = deli_next + 1              /* position after end of line           */
                 nextChar  = ""
              END
              ELSE
              DO
                 token_pos = deli_next + 1
                 nextChar  = SUBSTR( line, token_pos, 1 )     /* char *immediately* following */
              END

              resultArray[ resIndex ] = .array ~ of( tmpToken , "Q", nextChar, lineNr )
              resIndex = resIndex + 1   /* calculate next entry                         */
              ITERATE workLine          /* leave this loop, analyze rest of line        */
           END
        END


/**********************************************************************************/
        /* ---> extract symbols  */
        /* get first non-white-char */
        bIsSymbol = ( VERIFY( SUBSTR( line, token_pos, 1) , .orx.SymbolChars, "M" ) = 1 )
        IF bIsSymbol THEN                             /* o.k. a symbol token in hand  */
        DO
           CALL Pending_Comma
           deli_next = VERIFY( line, .orx.SymbolChars, "N", token_pos + 1 )   /* get next quote       */

           IF deli_next = 0 THEN                      /* o.k. only symbol left                        */
           DO
              deli_next = iLineLength + 1
           END

           /* extract symbol and return it    */
           tmpToken = SUBSTR( line, token_pos, deli_next - token_pos )
           IF deli_next - 1 = iLineLength THEN
           DO
              token_pos = deli_next                     /* position after end of line           */
              nextChar  = ""
           END
           ELSE
           DO
              token_pos = deli_next
              nextChar  = SUBSTR( line, token_pos, 1 )     /* char *immediately* following */
           END

           resultArray[ resIndex ] = .array ~ of( tmpToken , "S", nextChar, lineNr )
           resIndex = resIndex + 1   /* calculate next entry                         */
           ITERATE workLine          /* leave this loop, analyze rest of line        */
        END


/**********************************************************************************/
        /* ---> extract operators *and* comments ! */
        bIsOperator = ( VERIFY( SUBSTR( line, token_pos, 1) , .orx.OperatorChars, "M" ) = 1 )
        IF bIsOperator THEN                           /* o.k. an Operator token in hand                */
        DO
           deli_next = VERIFY( line, .orx.OperatorChars, "N", token_pos + 1 ) /* get next non-operator */

           IF deli_next = 0 THEN
           DO
              deli_next = iLineLength + 1
           END

           /* extract Operator and return it    */
           tmpToken = SUBSTR( line, token_pos, deli_next - token_pos )

           comPos = POS( "/*", tmpToken )       /* an opening comment found ?           */

      /******************** extract comment *********************************************/
           IF comPos > 0 THEN                   /* comment open found !                 */
           DO
              beforeCom = ""                    /* receives operators                   */

              DO WHILE comPos > 0

                 startLineNr = lineNr              /* remember line which started comment  */
                 countCom = 1
                 IF comPos > 1 THEN                /* operators immediately preceeding open comment        */
                 DO
                    beforeCom = beforeCom || SUBSTR( tmpToken, 1, comPos - 1 )  /* extract operators before comment */
                 END

                 comBegPos  = token_pos + comPos - 1       /* position on comment begin            */

                 token_pos  = token_pos + comPos           /* look *after* comment open            */


                 DO UNTIL countCom = 0             /* as long as comment active                    */
                    comPosOpen = POS( "/*", line, token_pos + 1 )  /* next open comment            */
                    comPosClose= POS( "*/", line, token_pos + 1 )  /* next close comment           */

                    SELECT
                       WHEN comPosOpen = 0 & comPosClose = 0 THEN  /* look in next line            */
                            DO
                               IF token_pos = 0 | countCom > 0 THEN
                                  tmpToken = SUBSTR(line, comBegPos )
                               ELSE
                                  tmpToken = SUBSTR(line, comBegPos, token_pos - comBegPos + 1 )

                               resultArray[ resIndex ] = .array ~ of(  tmpToken , "C", "", lineNr )
                               resIndex = resIndex + 1   /* calculate next entry                         */

                               lineNr   = lineNr + 1               /* get next line                */
                               IF lineNr > totLines THEN
                               DO
                                  RAISE SYNTAX 4.1 ARRAY( "<comment-not-closed-exception>" )
                               END

                               token_pos     = 0                   /* start from the beginning of the line */
                               line    = tmpArr[ lineNr ]          /* get next line to analyze             */
                               iLineLength = LENGTH( line )        /* get length of line                   */
                               comBegPos     = 1                   /* this line starts in comment "mode"   */
                            END

                       WHEN ( comPosClose > 0 & ( comPosOpen = 0 | comPosOpen > comPosClose) ) THEN
                            DO                                     /* a comment end found          */
                               countCom = countCom - 1
                               token_Pos = comPosClose + 1          /* leave comment close alone    */
                            END

                       WHEN comPosOpen > 0 THEN                    /* another comment open found   */
                            DO
                               countCom = countCom + 1
                               token_Pos = comPosOpen + 1           /* leave comment open alone     */
                            END

                       OTHERWISE
                            DO
                               say "--------------> not allowed; error !!!"
                            END
                    END
                 END

                 tmpToken = SUBSTR(line, comBegPos, comPosClose - comBegPos + 2 )

                 IF comPosClose + 1 = iLineLength THEN
                    nextChar = ""
                 ELSE
                    nextChar = SUBSTR( line, comPosClose + 2, 1 )

                 resultArray[ resIndex ] = .array ~ of( tmpToken , "CE", nextChar, lineNr )
                 resIndex = resIndex + 1   /* calculate next entry                         */

                 token_pos = token_pos + 1              /* position to reestablish analyzing    */

                 comPos = 0                             /* make sure loop is left by default        */
                 IF beforeCom <> "" THEN                /* comment started in middle of operators ? */
                 DO
                    bIsOperator = ( VERIFY( SUBSTR( line, token_pos, 1) , .orx.OperatorChars, "M" ) = 1 )
                    IF bIsOperator THEN                 /* o.k. operators follow immediately    */
                    DO
                      deli_next = VERIFY( line, .orx.OperatorChars, "N", token_pos + 1 ) /* get next non-operator */
                      IF deli_next = 0 THEN
                         deli_next = iLineLength + 1

                      /* extract Operator and return it    */
                      tmpToken = SUBSTR( line, token_pos, deli_next - token_pos )
                      comPos = POS( "/*", tmpToken )       /* an opening comment found ?           */

                      IF comPos = 0 THEN                        /* no comment open found !      */
                      DO
                         beforeCom = beforeCom || tmpToken
                         token_pos = deli_next
                      END
                    END
                 END
              END

              IF beforeCom <> "" THEN                   /* o.k. operators were before comment   */
              DO
                 resultArray[ resIndex ] = .array ~ of( beforeCom , "O", "/", lineNr )
                 resIndex = resIndex + 1   /* calculate next entry                         */
              END

              ITERATE workLine          /* leave this loop, analyze rest of line        */
           END
      /******************** extract operator ********************************************/
           ELSE                                 /* regular operator(s)                  */
           DO
              CALL Pending_Comma

              IF deli_next - 1 = iLineLength THEN
              DO
                 token_pos = deli_next          /* position after end of line           */
                 nextChar  = ""
              END
              ELSE token_pos = deli_next
              DO
                 token_pos = deli_next
                 nextChar  = SUBSTR( line, token_pos, 1 )     /* char *immediately* following */
              END

              resultArray[ resIndex ] = .array ~ of( tmpToken , "O", nextChar, lineNr )
              resIndex = resIndex + 1   /* calculate next entry                         */
           END

           ITERATE workLine          /* leave this loop, analyze rest of line        */
        END


/**********************************************************************************/
        /* o.k. now only ' ~ ( ) [ ] : ; , or white space left */
        tmpToken = SUBSTR( line, token_pos, 1 )
        CALL Pending_Comma

        IF token_pos = iLineLength THEN
        DO
           token_pos = token_pos + 1    /* position after end of line                   */
           nextChar  = ""
        END
        ELSE
        DO
           token_pos = token_pos + 1
           nextChar  = SUBSTR( line, token_pos, 1 )     /* char *immediately* following */
        END

        IF tmpToken = ";" THEN
        DO
           bNewLine  = .true
           resultArray[ resIndex - 1][ 3 ] = ""         /* delete nChar of preceding token */
           ITERATE workLine          /* leave this loop, analyze rest of line        */
        END

        IF  tmpToken = "," THEN
        DO
           bCommaPending = .true
           oldLineNr     = lineNr
           oldNextChar   = nextChar
           ITERATE workLine          /* leave this loop, analyze rest of line        */
        END

        resultArray[ resIndex ] = .array ~ of( tmpToken , "M", nextChar, lineNr )
        resIndex = resIndex + 1      /* calculate next entry                         */
     END
  END

/*
call sayerror "--- EOF --- encountered in line #" pp( lineNr - 1 )
*/

  resultArray[ resIndex ] = .array ~ of( "", "N", "", "" )     /* indicate new line    */
  resIndex = resIndex + 1   /* calculate next entry                         */

  RETURN resultArray


/* takes care of the case of a pending comma    */
Pending_Comma :
   IF \ bCommaPending THEN RETURN       /* no need to take care of comma*/

   /* pending comma is a true comma and not a line continuation, save it*/

   IF lineNr = oldLineNr THEN
   DO
      resultArray[ resIndex ] = .array ~ of( "," , "M", oldNextChar, lineNr )
      resIndex = resIndex + 1              /* calculate next entry         */
   END

   bCommaPending = .false               /* reset                        */
   RETURN



SYNTAX : RAISE PROPAGATE                /* raise SYNTAX error in caller */
/******************************************************************************/
/******************************************************************************/






/******************************************************************************/
/*                                                                            */
/* name:    Remove_Comments( array_REXX_Code )                                */
/*                                                                            */
/* purpose: remove comments from REXX-statements ...                          */
/*                                                                            */
/* returns: .array containing the REXX-code passed in via the array           */
/*          array_REXX_Code, which has all comments and empty lines removed,  */
/*          as well as joined lines (";") split, as well as split lines (",") */
/*          rejoined, as well as leading and trailing blanks removed          */
/*                                                                            */
/* remarks: ---                                                               */
/*                                                                            */
/*          rgf, 95-10-20                                                     */
/*
   strip comments from a given line, take care of directive spanning multiple lines
   produce an array, *not* containing any comments, joining split lines,
   splitting lines at ";"
*/
:: ROUTINE remove_comments PUBLIC
  USE ARG tmpArg                /* array: 1. argument file-array, 2. argument counter */

  newArray = .array ~ new       /* make a new array, containing code only */
  newIndex = 1                  /* index to enter cleaned lines */

/*
say
*/

  comO = "/*"                   /* open comment string */
  comE = "*/"                   /* close comment string */
  comLength = LENGTH( comO )

  comments = 0
  tmpString = ""
  pendingLine = 0               /* line to be continued on next line ? */
  DO i = 1 TO tmpArg ~ items

     line = STRIP( tmpArg[i] )  /* get line to strip comment off */
/*
say
say LEFT( "", 80, "=" )
say
*/

     old_open  = 0              /* indicate a new line in which counter for a closing comment was not decremented yet */
     old_close = 0              /* indicate a new line in which counter for a closing comment was not decremented yet */

     IF comments > 0 THEN       /* line belongs to an active comment */
     DO
        next_open   = get_next_open( comO, line, 1, "*" )

        IF next_open > 0 THEN   /* found a nesting comment */
           comments = comments + 1

        next_close  = POS( comE, line, 1 )

        IF next_open = 0 & next_close = 0 THEN ITERATE  /* read next line */

/*
say LEFT( "/", 50, "/" )
tmprgf1 = line
tmprgf2 = copies( " ", length( line ) )
if next_open > 0 then tmprgf2 = overlay("<", tmprgf2, next_open )
if next_close > 0 then tmprgf2 = overlay(">", tmprgf2, next_close )
say "comments > 0" "         line" pp( line ) pp( comments ) "line#" pp( i )
say "            " "comments open" pp( tmprgf2 ) "comments" pp( comments )
say LEFT( "\", 50, "\" )
*/


     END
     ELSE                       /* check whether comment on this line */
     DO
        IF \pendingLine THEN    /* only memorize the starting line, in case lines are split */
           presentLine = i      /* memorize starting line */

        commentPos = comment_available( line, comO, 1, presentLine ) /* is open comment enclosed in quotes ? */
        IF commentPos > 0 THEN
        DO
           next_open  = commentPos
           next_close = POS(comE, line, next_open + comLength )  /* get next closing comment */
           comments = 1

           IF next_open > 1 THEN
              tmpString = tmpString || SUBSTR( line, 1, next_open - 1)  /* save first non-comment part */
        END

/*
say LEFT( "/", 50, "/" )
tmprgf1 = line
tmprgf2 = copies( " ", length( line ) )
if next_open > 0 then tmprgf2 = overlay("<", tmprgf2, next_open )
if next_close > 0 then tmprgf2 = overlay(">", tmprgf2, next_close )
say "comments = 0" "         line" pp( line ) "line#" pp( i )
say "            " "no comm open?" pp( tmprgf2 ) "comments" pp( comments )
say "                               tmpString" pp( tmpString )
say LEFT( "\", 50, "\" )
*/

     END

     IF comments = 0 THEN
     DO
        tmpString = tmpString || line

        pendingLine = RIGHT( line, 1 ) = ","
        IF pendingLine THEN                     /* line continued */
        DO
           tmpString = OVERLAY(" ", tmpString, LENGTH( tmpString ) ) /* replace last comma with a blank */
           ITERATE
        END

        newIndex =  break_up( newArray, newIndex, tmpString, presentLine )/* save edited line */
        tmpString = ""
        ITERATE                                 /* read next line */
     END


     /* now work on comments ... */
     IF old_close = 0 & next_close > 0 THEN     /* decrease counter as this end of comment was detected in a freshly read line */
        comments = comments - 1

     IF next_close = 0 THEN             /* no close comment on this line, check for nested comments */
     DO
        DO WHILE next_open <> 0
           next_open   = get_next_open( comO, line, next_open + comLength, "*" )
           IF next_open > 0 THEN
              comments = comments + 1   /* found a nesting comment */
        END
     END
     ELSE                               /* a close comment found, check whether comment is being closed */
     DO
        old_close = next_close
        DO WHILE next_close > 0

           IF next_open > next_close THEN                    /* search additional closing comments */
           DO
/*
say "???b next_open > next_close" "next_open" pp( next_open ) "next_close" pp( next_close )
say "???b old_close" pp( old_close )
*/

              next_close = POS(comE, line, old_close + comLength )    /* get next closing comment */

              /*************************/
              IF next_close = 0 THEN            /* only one comE on the line, which is *before* next comO */
              DO
                 tmpString = tmpString || SUBSTR( line, old_close + comLength, next_open - old_close - comLength )
              END
              ELSE
              DO
              /*************************/

   /*
   say "???b new:"  "next_close" pp( next_close ) "next_open - next_close" pp( next_open - next_close )
   */

                 IF next_open - next_close < comLength THEN    /* comE-trap ! */
                 DO
   /*
   say "???b next_open - next_close < comLength" "next_open" pp( next_open ) "next_close" pp( next_close )
   */

                    next_close = POS(comE, line, next_open + comLength )
                    IF next_close > 0 THEN
                    DO
/*
   say "???b next_close > 0" "next_close" pp( next_close ) "old_close" pp( old_close )
*/
                       IF old_close > 0 THEN
                       DO
/*
   say "???b BINGO old_close > 0" "old_close" pp( old_close )
*/
                          tmpString = tmpString || SUBSTR( line, old_close + comLength, next_open - old_close - comLength )
                       END
                    END
                 END

                 IF next_close > 0 THEN            /* found another ending comment */
                 DO
                    comments = comments - 1
                    old_close = next_close
                 END

              /*************************/
              END
              /*************************/

           END
           ELSE IF next_open <> 0 & next_open < next_close THEN /* any opens in between ? */
           DO
              old_open = next_open
              next_open = get_next_open( comO, line, next_open + comLength , "*" )

              IF next_open > 0 THEN
              DO
                 comments = comments + 1
              END
           END
           ELSE                                                 /* search additional closing comments */
           DO
              next_close = POS(comE, line, old_close + comLength )    /* get next closing comment */
              IF next_close > 0 THEN            /* found another ending comment */
              DO
                 old_close = next_close
                 comments = comments - 1
              END
           END
/*
say LEFT( "/", 50, "/" )
tmprgf1 = line
tmprgf2 = copies( " ", length( line ) )
if next_open > 0 then tmprgf2 = overlay("<", tmprgf2, next_open )
if next_close > 0 then tmprgf2 = overlay(">", tmprgf2, next_close )
say "going furthe" "         line" pp( line ) "line#" pp( i )
say "            " "no comm open?" pp( tmprgf2 ) "comments" pp( comments )
say "                               tmpString" pp( tmpString )
say LEFT( "\", 50, "\" )
*/

           IF comments = 0 THEN         /* o.k. comment ended */
           DO
/*
say "???a comments = 0" "next_open" pp( next_open ) "old_close" pp( old_close )
*/
              old_open = next_open
              next_open = comment_available( line, comO, old_close + comLength, presentLine ) /* another opening comment on line ? */

              IF next_open > 0 THEN     /* possible interval without comment */
              DO

                 tmpLength = next_open - (old_close + comLength)

/*
say "???a next_open > 0 & comments = 0;" "next_open" pp( next_open) "old_close" pp( oldClose ) "tmpLength" pp( tmpLength)
*/

                 IF tmpLength > 0 THEN
                    tmpString = tmpString || SUBSTR( line, old_close + comLength, tmpLength  )

                 comments = 1
                 next_close = POS(comE, line, next_open + comLength )  /* get next closing comment */
                 IF next_close > 0 THEN
                 DO
                    old_close = next_close
                    comments = 0
                 END
              END
              ELSE                      /* up to end of line is without comments */
              DO
/*
say "???a BINGO, else..." "old_close + comLength" pp( old_close + comLength)
*/
                 tmpString = tmpString || SUBSTR( line, old_close + comLength )
                 LEAVE                     /* leave the loop, line processed */
              END
           END  /* comments = 0 */
        END
     END

/*
say LEFT( "/", 50, "/" )
tmprgf1 = line
tmprgf2 = copies( " ", length( line ) )
if next_open > 0 then tmprgf2 = overlay("<", tmprgf2, next_open )
if next_close > 0 then tmprgf2 = overlay(">", tmprgf2, next_close )
say "at bottom   " "         line" pp( line ) "line#" pp( i )
say "            " "no comm open?" pp( tmprgf2 ) "comments" pp( comments )
say "                               tmpString" pp( tmpString )
say LEFT( "\", 50, "\" )
*/

     IF comments = 0 & tmpString <> "" THEN               /* no comment active, write line so far */
     DO
        tmpString   = STRIP( tmpString )
        pendingLine = RIGHT( tmpString, 1 ) = ","

        IF pendingLine THEN                     /* line continued */
        DO
           tmpString = OVERLAY(" ", tmpString, LENGTH( tmpString) ) /* replace last comma with a blank */
        END
        ELSE                                    /* save edited line */
        DO
           newIndex =  break_up( newArray, newIndex, tmpString, presentLine )
           tmpString = ""
        END
     END
  END

  SELECT        /* check whether everything is o.k., leftover to be dealt with ? */
     WHEN comments > 0    THEN
                    CALL sayError "l#" pp( presentLine )":" PP( SUBWORD( tmpArg[ presentLine ], 1, 5) || "..." ),
                         "- error - opening comment never gets closed !"

     WHEN tmpString <> "" THEN newIndex =  break_up( newArray, newIndex, tmpString, presentLine )
     OTHERWISE NOP
  END

  RETURN newArray


/******************************************************************************/

/* routine to determine next open comment *within* an active comment, takes care of the case "*_/_*" */
:: ROUTINE get_next_open
  USE ARG comO, line, startPos, preceding

  IF \VAR( "startPos" )  THEN startPos  = 1       /* default */
  IF \VAR( "preceding" ) THEN preceding = ""      /* default */
  length = LENGTH( preceding )

  old_pos = 0
  tmp_open = startPos

/*
say
say "--- get_next_open() - entered, startPos" pp( startPos )
say "---" "line" pp( line )
say "---" "     " RIGHT( "^", startPos )
*/

  DO FOREVER
     tmp_open = POS( comO, line, tmp_open )     /* find an open comment */

     IF tmp_open = 0 THEN
do
/*
say "--- get_next_open() - exiting, tmp_open" pp( tmp_open )
*/
     RETURN 0              /* no begin comment found */
end

     IF preceding <> "" THEN    /* check wheter string immediately preceding comment open, invalidates it */
     DO
        /* check whether it is a true comment open, i.e. not "*_/_*" */
        IF tmp_open - length >= startPos THEN
        DO
/*
say "---" "?    " RIGHT( "^", tmp_open - length ) "extracted" pp( SUBSTR(line, tmp_open - length, length )  )
*/
           IF SUBSTR(line, tmp_open - length, length ) = preceding THEN
           DO
              /* a sequence of *_/_*_/_*, ie an '*' embedded between comment close and comment open */
              IF ( tmp_open = old_pos + 2 ) & old_pos <> 0 THEN
do
/*
say "---" "r0   " RIGHT( "^", tmp_open ) "tmp_open" pp( tmp_open ) "old_pos" pp( old_pos )
say "--- get_next_open() - EXITING [* / * / *] found !, tmp_open" pp( tmp_open )
*/
                 RETURN tmp_open
end

              old_pos = tmp_open
              tmp_open = tmp_open + 1
              ITERATE                           /* was a comment close, try to find another */
           END
        END
     END
/*
say "---" "r1   " RIGHT( "^", tmp_open )
say "--- get_next_open() - exiting, tmp_open" pp( tmp_open )
*/
     RETURN tmp_open                            /* return found position of next starting comment */
  END






/* routine to determine next open comment *within* an active comment, takes care of the case "*_/_*" */
:: ROUTINE get_next_open_ori
  USE ARG comO, line, startPos, preceding

  IF \VAR( "startPos" )  THEN startPos  = 1       /* default */
  IF \VAR( "preceding" ) THEN preceding = ""      /* default */
  length = LENGTH( preceding )

  old_pos = 0
  tmp_open = startPos

/*
say
say "--- get_next_open() - entered, startPos" pp( startPos )
say "---" "line" pp( line )
say "---" "     " RIGHT( "^", startPos )
*/

  DO FOREVER


     tmp_open = POS( comO, line, tmp_open )     /* find an open comment */

     IF tmp_open = 0 THEN
do
/*
say "--- get_next_open() - exiting, tmp_open" pp( tmp_open )
*/
     RETURN 0              /* no begin comment found */
end

     IF preceding <> "" THEN    /* check wheter string immediately preceding comment open, invalidates it */
     DO
        /* check whether it is a true comment open, i.e. not "*_/_*" */
        IF tmp_open - length >= startPos THEN
        DO
/*
say "---" "?    " RIGHT( "^", tmp_open - length )
*/
           IF SUBSTR(line, tmp_open - length, length ) = preceding THEN
           DO
              /* a sequence of *_/_*_/_*, ie an '*' embedded between comment close and comment open */
              IF tmp_open = old_pos + 2 THEN
do
/*
say "---" "r    " RIGHT( "^", tmp_open )
say "--- get_next_open() - EXITING [ * / * / * ] found !, tmp_open" pp( tmp_open )
*/
                 RETURN tmp_open
end

              old_pos = tmp_open
              tmp_open = tmp_open + 1
              ITERATE                           /* was a comment close, try to find another */
           END
        END
     END
/*
say "---" "r    " RIGHT( "^", tmp_open )
say "--- get_next_open() - exiting, tmp_open" pp( tmp_open )
*/
     RETURN tmp_open                            /* return found position of next starting comment */
  END



/******************************************************************************/

/* check whether a comment is available, i.e. a common open not within quotes */
:: ROUTINE comment_available
  USE ARG line, comO, start, presentLine

  IF \var( "comO"  ) THEN comO = "/*"
  IF \var( "start" ) THEN start = 1

  DO FOREVER                 /* handle case where comment leadin is within quotes */
     commentPos = POS( comO, line, start )              /* find a comment open string */

     IF commentPos = 0 THEN LEAVE                       /* no comment at this line */

     IF \token_in_string( line, commentPos, presentLine ) THEN LEAVE    /* true comment open */

     start = commentPos + 1                             /* look for next start of a comment */
  END
  RETURN commentPos
/******************************************************************************/


/* break into multiple lines, if line was concatenated with ";" */
:: ROUTINE break_up              PUBLIC
  USE ARG newArray, newIndex, tmpString, oldLine

  start = 1
  DO WHILE tmpString <> ""
     semi_pos = POS( ";", tmpString, start )
     IF semi_pos > 0 THEN
     DO
        IF token_in_string( tmpString, semi_pos, oldLine ) THEN   /* within quotes */
        DO
           start = semi_pos + 1                         /* look for another */
        END
        ELSE                                            /* split lines */
        DO
           til = semi_pos - 1                           /* get rid of semi-colon */
           PARSE VAR tmpString tmp +(til) +1 tmpString
           IF tmp <> "" THEN
           DO
              newArray[ newIndex, 1 ] = STRIP( tmp )
              newArray[ newIndex, 2 ] = oldLine
              newIndex = newIndex + 1

              start = 1
           END
        END
        ITERATE
     END

     /* no more semi-colons */
     newArray[ newIndex, 1 ] = STRIP( tmpString )
     newArray[ newIndex, 2 ] = oldLine
     newIndex = newIndex + 1
     LEAVE
  END

  RETURN newIndex
/******************************************************************************/







/******************************************************************************/
/* from orexx/samples   */
/* Define the stack class (a subclass of Object Class) */
::CLASS stack

/* Define the init method on the stack class */
::METHOD init
EXPOSE contents                             /* Object var with stack contents */

self ~ init:super                           /* Run init method of superclass  */
contents = .list~new                        /* Use a list to keep stack items */



/* Define the push method on the stack class */
::METHOD push
EXPOSE contents
USE ARG item

contents ~ insert(item, .nil)

/* Define the pop method on the stack class */
::METHOD pop
EXPOSE contents

IF contents ~ items > 0 THEN
  RETURN contents ~ remove(contents~first)
ELSE
  RETURN .nil

/* Define the items method on the stack class to return number of items */
::METHOD items
EXPOSE contents

RETURN contents~items






